import numpy as np
import matplotlib.pyplot as plt

import warnings
warnings.filterwarnings('ignore')

def plot_samples_states(samples, states, show=True):
    fig, ax = plt.subplots(nrows=2, ncols=1)
    ax[0].plot(samples, label='Sample returns', )
    ax[1].plot(states, label='Sampled true states', ls='dotted')
    plt.legend()

    if show is True:
        plt.show()
    return fig, ax

def plot_samples_states_viterbi(samples, viterbi_states, true_states, show=True):
    fig, ax = plt.subplots(nrows=2, ncols=1)
    ax[0].plot(samples, label='Sample returns')
    ax[1].plot(viterbi_states, label='Sampled viterbi states', ls='dotted')
    ax[1].plot(true_states, label='Sampled true states', ls='solid')
    plt.legend()

    if show == True:
        plt.show()
    return fig, ax

def plot_posteriors_states(posteriors, states, true_regimes, show=True):
    fig, ax = plt.subplots(nrows=2, ncols=1)
    ax[0].plot(posteriors[:, 0], label='Posteriors state 1', )
    ax[0].plot(posteriors[:, 1], label='Posteriors state 2', )
    ax[1].plot(states, label='Predicted states', ls='dotted')
    ax[1].plot(true_regimes, label='True states', ls='dashed')

    plt.legend()
    if show == True:
        plt.show()
    return fig, ax


def plot_2state_hmm_params(tpm, mu, std, true_tpm, true_mu, true_std):
    fig, ax = plt.subplots(nrows=3, ncols=2, figsize=(12, 7), sharex=True)
    ax[0, 0].plot(tpm[:, 0], label='P(1)', )
    ax[0, 0].plot([true_tpm[0, 0]] * len(tpm), label='True P1', ls='dashed')
    ax[0, 1].plot(tpm[:, 1], label='P(2)', )
    ax[0, 1].plot([true_tpm[1, 1]] * len(tpm), label='True P2', ls='dashed')
    # ax[0].legend()

    ax[1, 0].plot(mu[:, 0])
    ax[1, 0].plot([true_mu[0]] * len(mu), ls='dashed')
    ax[1, 1].plot(mu[:, 1])
    ax[1, 1].plot([true_mu[1]] * len(mu), ls='dashed')

    ax[2, 0].plot(std[:, 0])  # label='True states', ls='dashed')
    ax[2, 0].plot([true_std[0]] * len(std), ls='dashed')
    ax[2, 1].plot(std[:, 1])  # label='True states', ls='dashed')
    ax[2, 1].plot([true_std[1]] * len(std), ls='dashed')

    plt.show()


import warnings

import numpy as np
import pandas as pd
import tqdm

from hmmpy.finance.backtest import Backtester
from hmmpy.utils.data_prep import DataPrep


warnings.filterwarnings('ignore')
pd.set_option('display.max_columns', 10); pd.set_option('display.width', 320)
np.seterr(divide='ignore')

class BacktestAlternatives(Backtester):

    def __init__(self):
        super().__init__()

    def backtest_equal_weighted(self, df_rets, rebal_freq='M', port_val=1000, start_weights=None):
        """
       Backtest an equally weighted portfolio, with specified rebalancing frequency.

       Parameters
       ----------
       df_rets : DataFrame of shape (n_samples, n_assets)
           Historical returns for each asset i. Cash must be at the last column position.
       rebal_freq : int, default=20
            Rebalance frequency. Default is 20, i.e monthly.
       port_val : float, default=1000
           Starting portfolio value.
       start_weights : ndarray of shape (n_assets,)
           Current (known) portfolio weights at the start of backtest. Default is 100% allocation to cash.
           Cash must be the last column in df_rets.
       """
        self.port_val = np.array([0, port_val])
        self.n_assets = df_rets.shape[1]
        equal_weights = np.array([1 / self.n_assets] * self.n_assets)  # Vector of shape (n_assets,)

        if start_weights == None:  # Standard init with 100% allocated to cash
            start_weights = np.zeros(self.n_assets)
            start_weights[-1] = 1.
        else:
            start_weights = start_weights

        weights = start_weights
        trade_cost, turnover = [], []

        # Group data into months - average sample size is 20
        # Then for each month loop over the daily returns and update weights
        # The problem is recursive and thus requires looping done this way
        for month_dt, df_group in tqdm.tqdm(df_rets.groupby(pd.Grouper(freq=rebal_freq))):
            # Compute transaction costs for each month. Subtracted from gross ret the first of the month
            delta_weights = equal_weights - weights
            trans_cost = self.transaction_costs(delta_weights)
            weights = equal_weights  # Reset weights
            for day in range(len(df_group)):
                # Calculate gross returns for portfolio and append it
                if day == 0:
                    gross_ret = (1 + df_group.iloc[day]) * (1-trans_cost)
                else:
                    gross_ret = 1 + df_group.iloc[day]

                new_port_val = weights @ gross_ret * self.port_val[-1]
                self.port_val = np.append(self.port_val, new_port_val)

                new_w = gross_ret * weights
                new_w /= new_w.sum()  # Weights sum to 1
                weights = new_w  # Update weights each iteration

            trade_cost.append(trans_cost)
            turnover.append(np.linalg.norm(delta_weights, ord=1) / 2)  # Half L1 norm

        self.port_val = self.port_val[1:]  # Throw away first observation since it is artificially set to zero

        # Annualized average trading ost
        self.trans_cost = np.array(trade_cost)
        self.annual_trans_cost = 12 / len(self.trans_cost) * self.trans_cost.sum()

        # Compute average annualized portfolio turnover
        self.monthly_turnover = np.array(turnover)
        self.annual_turnover = 12 / len(self.monthly_turnover) * self.monthly_turnover.sum()


if __name__ == "__main__":
    data = DataPrep(out_of_sample=True)
    data.rets = data.rets.loc['2007-09-20':]

    backtest = BacktestAlternatives()

    backtest.backtest_equal_weighted(data.rets, rebal_freq='M')
    metrics = backtest.single_port_metric(data.prices, backtest.port_val, compare_assets=True)
    print(metrics)

    backtest.plot_port_val(data.prices, backtest.port_val, start=None, savefig=None)


import numpy as np
import pandas as pd; pd.set_option('display.max_columns', 10); pd.set_option('display.width', 320)
import cvxpy as cp

from hmmpy.utils.data_prep import load_returns


class MPC:
    """
    Solve the model predictive control problem using convex optimization.

    This class is strictly made for use with the CVXPY library. All methods are created to be
    compatible with this library and as such use of Numpy or Pandas is limited.

    Parameters
    ----------
    rets : ndarray of shape (n_preds, n_assets)
        Return predictions for each asset h time steps into the future.
    covariances : ndarray of shape (n_assets, n_assets)
        Covarince matrix of returns
    prev_port_vals : ndarray of dynamic length
        Times series of portfolio value at all previous time steps
    start_weights : ndarray of shape (n_assets,)
        Current (known) portfolio weights at time t.
    short_cons : str, default='LLO'
        How portfolio shorting is determined.
        'LLO': Leveraged-Long-Only, allowed to short risk-free asset.
        'long_only': No shorting allowed.
        'max_leverage': Shorting constraint entirely determined by max_leverage parameter
    max_leverage : float, default=2.0
        Maximum tolerable leverage defined as the sum of absolute values of the
        decimal weight in each asset apart from the risk-free asset.
    """

    def __init__(self, rets, covariances, prev_port_vals, start_weights,
                 max_drawdown=0.4, gamma_0=5, kappa1=0.004, kappa2=0.,
                 rho1=0., rho2=0.0005, rho_rf=0.001, max_holding=0.4, max_holding_rf=1., max_leverage=2.0,
                 short_cons='LO', rf_included=True,eps=0.0000001):
        if not short_cons in ['LLO', 'leveraged_long_only', 'LO', 'long_only', 'max_leverage', 'LS', None]:
            raise Exception("""short_cons must be declared as a str.
                            Options include 'LLO', 'leveraged_long_only', 'LO', 'long_only', 'max_leverage', 'LS, None.""")

        self.max_drawdown = max_drawdown
        self.gamma_0 = gamma_0
        self.kappa1 = kappa1
        self.kappa2 = kappa2
        self.rho1 = rho1
        self.rho2 = rho2
        self.rho_rf = rho_rf
        self.max_holding = max_holding
        self.max_holding_rf = max_holding_rf
        self.max_leverage = max_leverage
        self.short_cons = short_cons
        self.rf_included = rf_included
        self.eps = eps

        self.rets = np.array(rets)
        self.cov = np.array(covariances)
        self.prev_port_vals = prev_port_vals

        self.n_assets = self.rets.shape[1]
        self.n_preds = len(self.rets)
        self.start_weights = start_weights

    def single_period_objective_func(self, current_weights, prev_weights, rets, cov):
        """
        Compiles all individual components in the objective function into one.

        The method works for a single period. Remaining periods are bundled in cvxpy_solver method.

        Parameters
        ----------
        current_weights : CVXPY Variable of shape (n_assets,)
            Portfolio weights in the given time step t
        prev_weights : CVXPY Variable of shape (n_assets,)
            Portfolio weights in the previous time step t-1
        rets : ndarray of shape (n_assets,)
            Predicted returns for each asset at time t

        Returns
        -------
        objective : float
            Scalar value of objective function evaluated at time t.
        """
        port_ret = self._port_ret(current_weights, rets)
        trading_cost_ = self._trading_cost(current_weights, prev_weights)
        holding_cost_ = self._holding_cost(current_weights)
        port_risk = self._port_risk_control(current_weights, cov)

        objctive = port_ret - trading_cost_ - holding_cost_ - self.gamma * port_risk
        return objctive

    def single_period_constraints(self, current_weights):
        """
        Construct portfolio constraints for a single period. Bundled across time in cvxpy_solver method.

        Parameters
        ----------
        current_weights : CVXPY Variable of shape (n_assets,)
            Portfolio weights in the given time step t

        Returns
        -------
        constraints : list
            list containing all construct for time period t.
        """
        constraints = [cp.sum(current_weights) == 1, cp.abs(current_weights[:-1]) <= self.max_holding]
        constraints += [cp.abs(current_weights[-1]) <= self.max_holding_rf]  # Risk-free asset constraint

        # Leverage constraint. Exclude risk-free asset from this computation
        constraints += [cp.sum(cp.abs(current_weights[:-1])) <= self.max_leverage]
        if self.short_cons in ['LLO', 'leveraged_long_only']:  # Can only short risk-free asset
            constraints += [current_weights[:-1] >= 0]
        elif self.short_cons in ['LO', 'long_only']:
            constraints += [current_weights >= 0]
        elif self.short_cons in ['max_leverage', 'LS', None]:
            pass

        return constraints

    def cvxpy_solver(self, verbose=False):
        """
        Compile the optimization problem into CVXPY objects and solve.

        Returns optimal weights for each forecasted time period as ndarray of shape (n_preds, n_assets).
        """
        self.gamma = self._gamma_from_drawdown_control()

        objective = 0
        constr = []

        # weights variable depends on whether there is a risk-free asset in data
        if self.rf_included is True:
            # variable with shape h+1 predictions so first row
            # can be the known (non-variable) portfolio weight at time t
            weights = cp.Variable(shape=(self.n_preds + 1, self.n_assets))
        else:
            # Set rf to zero in all preds and cov
            self.rets = np.insert(self.rets, self.n_assets, 0, axis=1)
            self.cov = np.insert(self.cov, self.n_assets, 0, axis=-2)
            self.cov = np.insert(self.cov, self.n_assets, 0, axis=-1)  # Has to be done in two steps for cov due to dims
            self.start_weights = np.append(self.start_weights, 0)

            weights = cp.Variable(shape=(self.n_preds + 1, self.n_assets+1))
            rf_zero_weights = np.zeros(shape=self.n_preds)
            constr += [weights[1:, -1] == 0]  # Keep rf pos at zero since it non-present in this case


        # Loop through each row in the weights variable and construct the optimization problem
        # Note this loop is very cpu-light since no actual computations takes place inside it
        for t in range(1, weights.shape[0]):
            # sum problem objectives. Weights are shifted 1 period forward compared to self.rets
            # Concatenates objective and constraints in lists
            objective += self.single_period_objective_func(weights[t], weights[t-1], self.rets[t-1], self.cov[t-1])
            constr += self.single_period_constraints(weights[t])  # Concatenate constraints

        constr += [weights[0] == self.start_weights]  # first weights are fixed at known current portfolio

        prob = cp.Problem(cp.Maximize(objective), constr) # Construct maximization problem
        prob.solve(solver=cp.ECOS, verbose=verbose)

        if self.rf_included is True:
            opt_var = weights.value
        else:
            opt_var = weights.value[:, :-1]

        if verbose is True:
            print("Shape of var: ", opt_var.shape)
            temp_df = pd.DataFrame(opt_var).round(3)
            temp_df['sum_weights'] = np.sum(opt_var, axis=1)
            print(temp_df)

        return opt_var[1:]  # Discard first row which is not a variable.

    def _port_ret(self, weights, rets):
        """
        Compute portfolio return at a specific point in time.

        Takes a vector of weights and another vector of returns at time t.
        Returns the dot product -> Scalar.
        """
        port_ret_ = rets @ weights
        return port_ret_

    def _trading_cost(self, current_weights, prev_weights):
        """
        Using current and previous portfolio weights computes the trading costs as scalar value.

        Assumes no cost in trading risk-free asset and thus discards last value in delta_weight.
        """
        delta_weight = current_weights - prev_weights
        delta_weight = delta_weight[:-1]  # No costs associated with risk-free asset
        trading_cost = self.kappa1 * cp.abs(delta_weight) + self.kappa2 * cp.square(delta_weight)  # Vector of trading costs per asset

        return cp.sum(trading_cost)

    def _holding_cost(self, weights):
        """
        Portfolio holding costs.
        """
        weights_no_rf = weights[:-1]
        weights_rf = weights[-1]
        holding_cost_ = self.rho1 * cp.abs(weights_no_rf) + self.rho2 * cp.square(weights_no_rf) + self.rho_rf*cp.square(weights_rf)
        return cp.sum(holding_cost_)

    def _gamma_from_drawdown_control(self):
        """
        Compute gamma (risk-aversion parameter) using drawdown control.

        """
        # From all previous total portfolio values get the highest one
        previous_peak = np.max(self.prev_port_vals)

        drawdown_t = 1 - self.prev_port_vals[-1] / previous_peak
        denom = np.max([self.max_drawdown - drawdown_t, self.eps])  # If drawdown limit is breached use a number very close to zero
        gamma = self.gamma_0 * self.max_drawdown / denom

        return gamma

    def _port_risk_control(self, weights, cov):
        """
        Computes portfolio risk parameter. Currently set to portfolio variance.
        """
        port_var = cp.quad_form(weights, cov)  # CVXPY method for doing: w^T @ COV @ w
        return port_var

if __name__ == "__main__":
    df_ret = load_returns()
    cov = df_ret.cov().to_numpy()

    ret_pred = df_ret.iloc[-15:].to_numpy()
    weights = np.zeros(len(cov))
    weights[-1] = 1.
    prev_port_vals = np.array([100, 105,110,100,105,120,130,150,135,160])

    model = MPC(ret_pred, cov, prev_port_vals)
    weigths = model.cvxpy_solver()





import numpy as np
from scipy import stats
import matplotlib.pyplot as plt


def simulate_2state_gaussian(N=200, means=None, std=None, plotting=False, random_state=42):
    '''Simulate data for bull and bear market using Normal distribution

    If means and std = None, the distributions follow those estimated by Hardy(2001)
    which are based on monthly returns from a stock index.
    '''

    if means == None:
        bull_mean = 0.1
        bear_mean = -0.05

    if std == None:
        bull_std = 0.1
        bear_std = 0.2

    np.random.seed(random_state)
    market_bull_1 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)
    market_bear_2 = stats.norm.rvs(loc=bear_mean, scale=bear_std, size=N)
    market_bull_3 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)
    market_bear_4 = stats.norm.rvs(loc=bear_mean, scale=bear_std, size=N)
    market_bull_5 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)

    # Create the list of true regime states and full returns list
    returns = np.array([market_bull_1] + [market_bear_2] + [market_bull_3] + [market_bear_4] + [market_bull_5]).flatten()
    true_regimes = np.array([np.zeros(N), np.ones(N), np.zeros(N), np.ones(N), np.zeros(N)]).flatten().astype(int)

    #true_regimes = np.array([len(market_bull_1)] + [len(market_bear_2)] + [len(market_bull_3)] + [len(market_bear_4)] + [len(market_bull_5)]).flatten()

    if plotting:
        plt.plot(returns)
        plt.plot(true_regimes)
        plt.show()

    return returns, true_regimes




def simulate_bear_t(N=200, means=None, std=None, df=None, plotting=True, random_state=42):  ### Simulate 3 state.
        '''Simulate data for bull and bear market using Normal distribution

            If means and std = None, the distributions follow those estimated by Hardy(2001)
            which are based on monthly returns from a stock index.
        '''
        if means == None:
            bull_mean = 0.1
            bear_mean = -0.05

        if std == None:
            bull_std = 0.1
            bear_std = 0.2

        if df == None:
            bear_df = N - 1  # Bull state is only normal distribution hence it is not specified.

        np.random.seed(random_state)
        market_bull_1 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)
        market_bear_2 = stats.t.rvs(loc=bear_mean, scale=bear_std, size=N, df=bear_df)
        market_bull_3 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)
        market_bear_4 = stats.t.rvs(loc=bear_mean, scale=bear_std, size=N, df=bear_df)
        market_bull_5 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)

        returns = np.array([market_bull_1] + [market_bear_2] + [market_bull_3] + [market_bear_4] + [market_bull_5]).flatten()
        true_regimes = np.array([np.zeros(N), np.ones(N), np.zeros(N), np.ones(N), np.zeros(N)]).flatten()

        if plotting:
            plt.plot(returns)
            plt.plot(true_regimes)
            plt.show()

        return returns, true_regimes


def simulate_3state_gaussian_t(N=500, means=None, std=None, df=None, plotting=True, random_state=42):  # Simulate 3 state.
    if means == None:
        bull_mean = 0.1
        bear_mean = -0.05
        recession_mean = -0.10

    if std == None:
        bull_std = 0.1
        bear_std = 0.2
        recession_std = 0.3

    if df == None:
        recession_df = N - 1  # Bull and bear state is normal distribution hence it is not specified.

    np.random.seed(random_state)
    market_bull_1 = stats.norm.rvs(loc= bull_mean, scale=bull_std, size=N)
    market_bear_2 = stats.norm.rvs(loc= bear_mean, scale=bear_std, size=N)
    market_recession_3 = stats.t.rvs(loc= recession_mean, scale=recession_std, size=N, df = recession_df)
    market_bull_4 = stats.norm.rvs(loc=bull_mean, scale=bull_std, size=N)
    market_recession_5 = stats.t.rvs(loc = recession_mean, scale = recession_std, size=N, df=recession_df)
    market_bear_6 = stats.norm.rvs(loc = bear_mean, scale = bear_std, size=N)
    market_bull_7 = stats.norm.rvs(loc = bull_mean, scale = bull_std, size=N)

    returns = np.array([market_bull_1] + [market_bear_2] + [market_recession_3] + [market_bull_4] + [market_recession_5] + [market_bear_6] + [market_bull_7]).flatten()
    true_regimes = np.array([np.zeros(N), np.ones(N), np.ones(N)*2, np.zeros(N), np.ones(N)*2, np.ones(N), np.zeros(N)]).flatten()

    if plotting:
        plt.plot(returns)
        plt.plot(true_regimes)
        plt.show()

    return returns, true_regimes


if __name__ == '__main__':
    returns, true_regimes = simulate_3state_gaussian_t(plotting= True)
    print(true_regimes)






import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

if __name__ == '__main__':
    df = pd.read_csv('../../data/price_series_old.csv', index_col='Time')
    df.index = pd.to_datetime(df.index)

    df1 = pd.read_excel('../../data/BarCap_USCorpHY_10yearspread.xlsx')
    df1 = df1.iloc[2:]
    df1.rename({'ticker': 'Time'}, inplace=True, axis=1)
    df1.set_index('Time', inplace=True)
    df1.index = pd.to_datetime(df1.index)


    df_merged = pd.merge(left=df, right=df1, how='left',
                         left_index=True, right_index=True)

    # Rearrange columns so rf is the last column
    cols = ['Hedge Funds Global',
             'MSCI World',
             'MSCI Emerging Markets',
             'DAX ',
             'European Public Real Estate',
             'S&P 500 ',
             'Oil',
             'Gold index',
             'Barclays US Treasury',
             'Private Equity',
             'CSI BARC Index',
             'T-bills rf']

    df_merged = df_merged[cols]
    df_merged.to_csv('../../data/price_series.csv')

    print(df_merged)


import numpy as np
from numpy import ndarray
from scipy.special import logsumexp

from hmmpy.base import BaseHiddenMarkov
from hmmpy.sampler import SampleHMM


class MLEHMM(BaseHiddenMarkov):
    """"
    Class for computing HMM's using the EM algorithm.

    Can be used to fit HMM parameters or to decode hidden states.

    Parameters
    ----------
    n_states : int, default=2
        Number of hidden states
    max_iter : int, default=100
        Maximum number of iterations to perform during expectation-maximization
    tol : float, default=1e-6
        Criterion for early stopping
    epochs : int, default=1
        Number of complete passes through the data to improve fit
    random_state : int, default = 42
        Parameter set to recreate output
    init: str
        Set to 'random' for random initialization.
        Set to None for deterministic init.

    Attributes
    ----------
    mu : ndarray of shape (n_states,)
        Fitted means for each state
    std : ndarray of shape (n_states,)
        Fitted std for each state
    tpm : ndarray of shape (n_states, n_states)
        Matrix of transition probabilities between states
    start_proba : ndarray of shape (n_states,)
        Initial state occupation distribution
    gamma : ndarray of shape (n_states,)
        Entails the probability of being in a state at time t knowing
        all the observations that has come and all the observations to come.
    """

    def __init__(self, n_states: int = 2, init: str = 'random', max_iter: int = 100, tol: float = 1e-6,
                 epochs: int = 10, random_state: int = 42):
        super().__init__(n_states, init, max_iter, tol, epochs, random_state)
        self.type = 'mle'

    def compute_log_posteriors(self, log_alphas, log_betas):
        """
        Expectation of being in state j at time t given observations, P(S_t = j | x^T).

        Parameters
        ----------
        log_alphas
        log_betas

        Returns
        -------
        gamma : ndarray of shape (n_samples, n_states)
            Expectation of being in state j at time t given observations, P(S_t = j | x^T)
        """
        log_gamma = log_alphas + log_betas
        normalizer = logsumexp(log_gamma, axis=1, keepdims=True)
        log_gamma -= normalizer
        return log_gamma

    def compute_log_xi(self, log_alphas, log_betas):
        """Expected number of transitions from state i to j, (P(S_t-1 = j, S_t = i | x^T)"""
        # Initialize matrix of shape j X j
        # Number of expected transitions from state i to j
        log_xi = np.zeros(shape=(len(log_alphas)-1, self.n_states, self.n_states))

        with np.errstate(divide='ignore'):
            log_tpm = np.log(self.tpm)

        for i in range(self.n_states):
            for j in range(self.n_states):
                log_xi[:, i, j] = log_tpm[i, j] + log_alphas[:-1, i] + \
                                 log_betas[1:, j] + self.log_emission_probs_[1:, j]

        normalizer = logsumexp(log_xi, axis=(1,2))  # Take log sum over last two axes and keep first one.
        log_xi = log_xi - normalizer[:, np.newaxis, np.newaxis]
        return log_xi

    def _e_step(self, X: ndarray):
        '''
        Do a single e-step in Baum-Welch algorithm (Derives Xi and Gamma w.r.t. traditional HMM syntax)
        '''
        self.emission_probs(X)
        llk, log_alphas = self._log_forward_proba()
        log_betas = self._log_backward_proba()

        log_gamma = self.compute_log_posteriors(log_alphas, log_betas)  # Shape (n_samples, n_states)
        log_xi = self.compute_log_xi(log_alphas, log_betas)  # Shape (n_samples, n_states, n_states)

        """
        # Initialize matrix of shape j X j
        # Number of expected transitions from state i to j
        xi = np.zeros(shape=(self.n_states, self.n_states))
        for j in range(self.n_states):
            for k in range(self.n_states):
                xi[j, k] = self.tpm[j, k] * np.sum(
                    np.exp(log_alphas[:-1, j] + log_betas[1:, k] + self.log_emission_probs_[1:, k] - llk))
        """

        return log_gamma, log_xi, llk

    def _m_step(self, X: ndarray, log_gamma, log_xi):
        '''
        Given u and f do an m-step.
        Updates the model parameters delta, Transition matrix and state dependent distributions.
         '''
        # Update transition matrix and initial probs
        #self.tpm = xi / np.sum(xi, axis=1).reshape((-1, 1))  # Check if this actually sums correct and to 1 on rows

        trans_probs = logsumexp(log_xi, axis=0) - logsumexp(log_gamma, axis=0).reshape(-1, 1)
        trans_probs = np.exp(trans_probs)
        self.tpm = trans_probs / np.sum(trans_probs, axis=1).reshape(-1,1)  # Make rows sum to 1

        gamma = np.exp(log_gamma)

        self.start_proba = gamma[0, :] / np.sum(gamma[0, :])

        # Update state-dependent distributions
        for j in range(self.n_states):
            self.mu[j] = np.sum(gamma[:, j] * X) / np.sum(gamma[:, j])
            self.std[j] = np.sqrt(np.sum(gamma[:, j] * np.square(X - self.mu[j])) / np.sum(gamma[:, j]))

    def _fit(self, X: ndarray, verbose=False):
        """ Container for looping part of fit-method"""
        self.old_llk = -np.inf  # Used to check model convergence
        self.best_epoch = -np.inf

        for epoch in range(self.epochs):
            # Do new init at each epoch
            self._init_params(X, output_hmm_params=True)

            for iter in range(self.max_iter):
                # Do e- and m-step
                log_gamma, log_xi, llk = self._e_step(X)
                self._m_step(X, log_gamma, log_xi)  # Updates mu, std and tpm

                # Check convergence criterion
                crit = np.abs(llk - self.old_llk)  # Improvement in log likelihood
                if crit < self.tol:
                    self.is_fitted = True
                    if llk > self.best_epoch:
                        self.best_epoch = llk
                        self.stationary_dist = self.get_stationary_dist(tpm=self.tpm)

                        self.best_tpm = self.tpm
                        self.best_start_proba = self.start_proba
                        self.best_mu = self.mu
                        self.best_std = self.std

                        if verbose == 2:
                            print(
                                f'Iteration {iter} - LLK {llk} - Means: {self.mu} - STD {self.std} - TPM {np.diag(self.tpm)} - Delta {self.start_proba}')
                    break
                else:
                    self.old_llk = llk

    def fit(self, X: ndarray, sort_state_seq=True, verbose=False):
        """
        Function iterates through the e-step and the m-step recursively to find the optimal model parameters.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Time series of data
        verbose : boolean
            False / True for extra information regarding the function.

        Returns
        ----------
        Derives the optimal model parameters
        """
        # Init parameters initial distribution, transition matrix and state-dependent distributions
        self.is_fitted = False
        self._fit(X, verbose=verbose)

        if self.is_fitted is False:
            max_iter = self.max_iter
            self.max_iter = max_iter * 2  # Double amount of iterations
            self._fit(X)  # Try fitting again
            self.max_iter = max_iter  # Reset max_iter back to user-input
            if self.is_fitted == False and verbose == True:
                print(f'MLE NOT FITTED -- epochs {self.epochs} -- iters {self.max_iter*2} -- mu {self.mu} -- std {self.std} -- tpm {np.diag(self.tpm)}')

        if sort_state_seq is True:
            self._check_state_sort()  # Ensures low variance state is first

        self.tpm = self.best_tpm
        self.start_proba = self.best_start_proba
        self.mu = self.best_mu
        self.std = self.best_std

    def _check_state_sort(self):
        # Sort array ascending and check if order is changed
        # If the order is changed then states are reversed
        if np.sort(self.best_std)[0] != self.best_std[0]:
            # TODO only works for 2-states
            self.best_mu = self.best_mu[::-1]
            self.best_std = self.best_std[::-1]
            self.best_tpm = np.flip(self.best_tpm)
            self.best_start_proba = self.start_proba[::-1]


if __name__ == '__main__':
    sampler = SampleHMM(n_states=2, random_state=42)
    X, viterbi_states, true_states = sampler.sample_with_viterbi(1000, 1)
    model = MLEHMM(n_states=2, init="random", random_state=42, epochs=20, max_iter=100)

    model.fit(X, verbose=True)
    print(model.tpm.sum(axis=1))

    #pool = Pool()
    #mapfunc = partial(model.fit, verbose=False)
    #result = pool.map(mapfunc, [X]*20)


import numpy as np
from scipy import stats
from scipy.special import digamma
import scipy.optimize as opt
import matplotlib.pyplot as plt

from hmmpy.utils.simulate_returns import simulate_2state_gaussian
from mle import MLEHMM



class EMTHiddenMarkov(MLEHMM):
    """
    Class for estimating HMMs with a mixture of gaussian and students t distributions.
    """

    def __init__(self, n_states: int = 2, init: str = 'random', max_iter: int = 100, tol: int = 1e-6,
                 epochs: int = 10, random_state: int = 42):
        super().__init__(n_states, init, max_iter, tol, epochs, random_state)
        self.dof = 2  # TODO how do we init this???

    def emission_probs(self, X: list):
        """ Compute all different log probabilities log(p(x)) given an observation sequence and n states

        Returns: T X N matrix
        """
        T = len(X)
        log_probs = np.zeros((T, self.n_states))  # Init N X M matrix

        # For all states evaluate the density function
        # In the last state, i.e. the high variance state use a t-dist
        for j in range(self.n_states):
            if j < (self.n_states-1):
                log_probs[:, j] = stats.norm.logpdf(X, loc=self.mu[j], scale=self.std[j])
            elif j == (self.n_states-1):
                log_probs[:, j] = stats.t.logpdf(X, loc=self.mu[j], scale=self.std[j], df=self.dof)

        probs = np.exp(log_probs)

        self.emission_probs_ = probs
        self.log_emission_probs_ = log_probs

        return probs, log_probs

    def _m_step(self, X, log_gamma, log_xi, iterations: int = 2):
        '''
        Given log_gamma and log_xi do an m-step.

        Update degrees of freedom iteratively.
        Updates the model parameters delta, Transition matrix and state dependent distributions.
         '''

        # init u_it
        u_it = np.square(self.std[-1]) * (self.dof + 1) / ((np.square(self.std[-1]) * self.dof) + np.square(X - self.mu[-1]))

        # Update transition matrix and initial probs
        self.T = log_xi / np.sum(log_xi, axis=1).reshape((-1, 1))  # Check if this actually sums correct and to 1 on rows
        self.delta = log_gamma[0, :] / np.sum(log_gamma[0, :])

        # TODO Remove iterations from findinf DOF as it is not necessary.
        # Update state-dependent distributions
        for iteration in range(iterations): # Iterate over this procedure until dof has converged
            for j in range(self.n_states):
                if j < (self.n_states-1):
                    self.mu[j] = np.sum(log_gamma[:, j] * X) / np.sum(log_gamma[:, j])
                    self.std[j] = np.sqrt(np.sum(log_gamma[:, j] * np.square(X - self.mu[j])) / np.sum(log_gamma[:, j]))

                if j == (self.n_states-1):
                    self.mu[j] = np.sum(log_gamma[:, j] * X * u_it) / (np.sum(log_gamma[:, j] * u_it))
                    self.std[j] = np.sqrt(np.sum(log_gamma[:, j] * np.square(X - self.mu[j])) / np.sum(log_gamma[:, j]))
                    self.u_it = np.square(self.std[j]) * (self.dof + 1) / ((np.square(self.std[j]) * self.dof) + np.square(X - self.mu[j]))

                    # Find root of some estimator function based on digammas and dof
                    def dof_estimator(dof):  # TODO research why Nystrup master p. 94 use exp(dof) rather than dof
                        term1 = 1 - digamma(0.5 * dof)
                        term2 = np.log(0.5 * dof)
                        term3 = digamma((dof + 1) / 2)
                        term4 = -np.log((dof + 1) / 2)
                        term5 = 1 / np.sum(log_gamma[:, j]) * np.sum(log_gamma[:, j] * (np.log(u_it) - u_it))
                        return term1 + term2 + term3 + term4 + term5

                    self.dof = opt.root(fun=dof_estimator, x0=self.dof).x




if __name__ == '__main__':
    model = EMTHiddenMarkov(n_states=2, epochs=5)
    returns, true_regimes = simulate_2state_gaussian(plotting=False)  # Simulate some X in two states from normal distributions

    model.fit(returns, verbose=0)
    states = model.decode(returns)

    plotting = False
    if plotting == True:
        fig, ax = plt.subplots(nrows=2, ncols=1)
        ax[0].plot(posteriors[:, 0], label='Posteriors state 1', )
        ax[0].plot(posteriors[:, 1], label='Posteriors state 2', )
        ax[1].plot(states, label='Predicted states', ls='dotted')
        ax[1].plot(true_regimes, label='True states', ls='dashed')

        plt.legend()
        plt.show()



import numpy as np
import tqdm
from scipy import stats

from hmmpy.base import BaseHiddenMarkov

class SampleHMM(BaseHiddenMarkov):
    """
    Class to handle sampling from HMM hidden_markov with predefined parameters.
    

    Parameters
    ----------
    n_states : int, default=2
        Number of hidden states
    hmm_params: dict
        hmm model parameters to sample from.
        To set params, create a dict with 'mu', 'std' and 'tpm' as kwds
        and their values in lists or ndarrays.
    random_state : int, default = 42
        Parameter set to recreate output
 
    Attributes
    ----------
    mu : ndarray of shape (n_states,)
        Fitted means for each state
    std : ndarray of shape (n_states,)
        Fitted std for each state
    tpm : ndarray of shape (n_states, n_states)
        Transition probability matrix between states
    """
    
    def __init__(self, n_states=2, frequency='daily', hmm_params=None, random_state=42):

        if hmm_params == None and frequency == "daily":  # hmm params following Hardy (2001)
            # Convert from monthly time scale t=20 to daily t=1
            hmm_params = {'mu': np.array([0.0123, -0.0157]) / 20,
                          'std': np.array([0.0347, 0.0778]) /np.sqrt(20),
                          'tpm': np.array([[1-0.0021, 0.0021],  # TODO figure out powers of vectors in python
                                           [0.0120, 1-0.0120]])
                          }
        elif hmm_params == None and frequency == "monthly":
            hmm_params = {'mu': np.array([0.0123, -0.0157]),
                          'std': np.array([0.0347, 0.0778]),
                          'tpm': np.array([[0.9629, 0.0371],
                                           [0.2101, 0.7899]])
                          }

        self.type = 'sampler'
        self.is_fitted = True
        self.n_states = n_states
        self.mu = np.array(hmm_params['mu'])
        self.std = np.array(hmm_params['std'])
        self.tpm = np.array(hmm_params['tpm'])
        self.stationary_dist = super().get_stationary_dist(self.tpm)
        self.start_proba = self.stationary_dist

        self.random_state = random_state
        np.random.seed(self.random_state)

    def sample(self, n_samples, n_sequences=1):
        '''
        Sample states from a fitted Hidden Markov Model.

        Parameters
        ----------
        n_samples : int
            Amount of samples to generate
        n_sequences : int, default=1
            Number of independent sequences to sample from, e.g. if n_samples=100 and n_sequences=3
            then 3 different sequences of length 100 are sampled

        Returns
        -------
        samples : ndarray of shape (n_samples, n_sequences)
            Outputs the generated samples of size n_samples
        sample_states : ndarray of shape (n_samples, n_sequences)
            Outputs sampled states
        '''
        mu = self.mu
        std = self.std
        tpm = self.tpm
        stationary_dist = self.stationary_dist

        state_index = np.arange(start=0, stop=self.n_states, step=1, dtype=np.int32)  # Array of possible states
        sample_states = np.zeros(shape=(n_samples, n_sequences), dtype=np.int32) # Init sample vector
        samples = np.zeros(shape=(n_samples, n_sequences))  # Init sample vector

        print(f'Simulating {n_sequences} of lengths {n_samples}')
        for seq in tqdm.tqdm(range(n_sequences)):
            sample_states[0, seq] = np.random.choice(a=state_index, size=1, p=stationary_dist)

            for t in range(1, n_samples):
                # Each new state is chosen using the transition probs corresponding to the previous state sojourn.
                sample_states[t, seq] = np.random.choice(a=state_index, size=1, p=tpm[sample_states[t - 1, seq], :])

            samples[:, seq] = stats.norm.rvs(loc=mu[sample_states[:, seq]], scale=std[sample_states[:, seq]], size=n_samples)

        if n_sequences == 1:
            sample_states = sample_states[:, 0]
            samples = samples[:, 0]

        return samples, sample_states

    def sample_t(self, n_samples, n_sequences=1, dof=5):
        '''
        Sample states from a fitted Hidden Markov Model.

        Parameters
        ----------
        n_samples : int
            Amount of samples to generate
        n_sequences : int, default=1
            Number of independent sequences to sample from, e.g. if n_samples=100 and n_sequences=3
            then 3 different sequences of length 100 are sampled

        Returns
        -------
        samples : ndarray of shape (n_samples, n_sequences)
            Outputs the generated samples of size n_samples
        sample_states : ndarray of shape (n_samples, n_sequences)
            Outputs sampled states
        '''
        mu = self.mu
        std = self.std
        tpm = self.tpm
        stationary_dist = self.stationary_dist

        state_index = np.arange(start=0, stop=self.n_states, step=1, dtype=np.int32)  # Array of possible states
        sample_states = np.zeros(shape=(n_samples, n_sequences), dtype=np.int32) # Init sample vector
        samples = np.zeros(shape=(n_samples, n_sequences))  # Init sample vector

        for seq in tqdm.tqdm(range(n_sequences)):
            sample_states[0, seq] = np.random.choice(a=state_index, size=1, p=stationary_dist)

            for t in range(1, n_samples):
                # Each new state is chosen using the transition probs corresponding to the previous state sojourn.
                sample_states[t, seq] = np.random.choice(a=state_index, size=1, p=tpm[sample_states[t - 1, seq], :])

            samples[:, seq] = stats.t.rvs(loc=mu[sample_states[:, seq]],
                                          scale=std[sample_states[:, seq]]/np.sqrt(dof/(dof-2)),
                                          size=n_samples, df=dof)

        if n_sequences == 1:
            sample_states = sample_states[:, 0]
            samples = samples[:, 0]

        return samples, sample_states

    def sample_with_viterbi(self, n_samples, n_sequences=1):
        samples, true_states = self.sample(n_samples, n_sequences)

        viterbi_states = np.empty(shape=(n_samples, n_sequences), dtype=float)
        if n_sequences == 1:
            viterbi_states = self.decode(samples)
        else:
            for i in range(n_sequences):
                viterbi_states[:, i] = self.decode(samples[:, i])

        return samples, viterbi_states, true_states


if __name__ == "__main__":
    model = SampleHMM(n_states=2)
    print(model.mu)
    print(model.std)
    print(model.tpm)
    print(model.stationary_dist)



    n_samples = 1000
    n_sequences = 1000
    X, viterbi_states, true_states = model.sample_with_viterbi(n_samples, n_sequences)


""" TODO
In Backtester.get_asset_dist when only 1 state is present the other takes on mean
values of 0 and covariances of 0. Is this a good result ?

Sometimes a state is not vistited or only has 1 observation in which case covariance matrix will return null results.
We need a solution for this; currently setting it equal to zero in those cases.

Create some way to do in-sample crossvalidation for hyperparameters

SHRINKAGE: Sort portfolios according to variance
    EVMA OR Hyperbolic
    Eksponentielt vægtet


"""
import warnings

import numpy as np
import pandas as pd
import tqdm
from matplotlib import pyplot as plt

from hmmpy.mle import MLEHMM
from hmmpy.finance.mpc_model import MPC
from hmmpy.utils.data_prep import DataPrep

warnings.filterwarnings('ignore')
pd.set_option('display.max_columns', 10); pd.set_option('display.width', 320)
np.seterr(divide='ignore')

class FinanceHMM:
    """
    Class to compute multivariate mixture distributions from n_assets based on a given HMM.

    Computes posteriors, state sequences as well as expected and forecasted returns and standard deviations.
    Transforms lognormal multivariate distributions into normal distributions and combines them into mixtures.

    Parameters
    ----------
    X : ndarray of shape (n_samples,)
        Times series data used to train the HMM.
    df : DataFrame of shape (n_samples, n_assets)
        Times series data used when estimating expected returns and covariances.
    model : hidden markov model
        Hidden Markov Model object.

    Attributes
    ----------
    preds : ndarray of shape (n_samples-window_len, n_preds, n_assets)
        mean predictions for each asset h time steps into the future at each time t.
    cov : ndarray of shape(n_samples-window_len, n_preds, n_assets, n_assets)
        predicted covariance matrix h time steps into the future at each time t.
    """

    def __init__(self, model):
        self.model = model
        self.n_states = model.n_states
        self.n_assets = None

    def get_cond_asset_dist(self, df, state_sequence):
        """
        Compute conditional multivariate normal distribution of all assets in each state.

        Assumes returns follow a multivariate log-normal distribution. Proceeds by first
        getting the conditional log of means and covariances and then transforming them
        back into normal varibles.

        Parameters
        ----------
        df : DataFrame of shape (n_samples, n_assets)
            log-returns for assets
        state_sequence : ndarray of shape (n_samples,)
            Decoded state sequence

        Returns
        -------
        mu : ndarray of shape (n_states, n_assets)
            Conditional mean value of each assets
        cov : ndarray of shape (n_states, n_assets, n_assets)
            Conditional covariance matrix
        """
        self.n_assets = df.shape[1]
        df = df.iloc[-len(state_sequence):]
        df['state_sequence'] = state_sequence
        groupby_state = df.groupby('state_sequence')
        log_mu, log_cov = groupby_state.mean(), groupby_state.cov()
        state_count = groupby_state.count().max(axis=1)  # Num obs in each state

        mu = np.zeros(shape=(self.n_states, self.n_assets))
        cov = np.zeros(shape=(self.n_states, self.n_assets, self.n_assets))

        # Loop through n_states present in current sample
        for s in log_mu.index:
            if state_count[s] > 1:  # If state_count not >1, covariance will return NaN
                mu[s], cov[s] = self.logcov_to_cov(log_mu.loc[s], log_cov.loc[s])

        return mu, cov

    def get_uncond_asset_dist(self, posteriors, cond_mu, cond_cov):
        """
        Compute unconditional multivariate normal distribution of all assets.

        Parameters
        ----------
        posteriors: ndarray of shape (n_preds, n_states)
            predicted posterior probability of being in state i at time t+h.
        cond_mu : ndarray of shape (n_states, n_assets)
            Conditional mean value of each assets
        cond_cov : ndarray of shape (n_states, n_assets, n_assets)
            Conditional covariance matrix
        Returns
        -------
        pred_mu : ndarray of shape (n_preds, n_assets)
            Conditional mean value of each assets
        pred_cov : ndarray of shape (n_preds, n_assets, n_assets)
            Conditional covariance matrix
        """
        pred_mu = np.inner(cond_mu.T, posteriors).T  # shape (n_preds, n_assets)

        cov_x1 = np.inner(posteriors, cond_cov.T)  # shape (n_preds, n_assets, n_assets)
        cov_x2 = pred_mu - cond_mu[:, np.newaxis]  # shape (n_states, n_preds)
        cov_x3 = np.einsum('ijk,ijk->ij', cov_x2, cov_x2)  # Equal to np.sum(X**2, axis=-1)
        cov_x4 = np.einsum('ij,ij->i', cov_x3.T, posteriors)  # Equal to np.sum(X3*posteriors, axis=1)
        pred_cov = cov_x1 + cov_x4[:, np.newaxis, np.newaxis]  # shape (n_preds, n_assets, n_assets)

        return pred_mu, pred_cov

    @staticmethod
    def logcov_to_cov(log_mu, log_cov):
        """
        Transforms log returns' means and covariances back into regular formats.

        Parameters
        ----------
        log_mu : DataFrame of shape (n_assets,)
        log_cov : DataFrame of shape (n_assets, n_assets)

        Returns
        -------
        mu : ndarray of shape (n_assets)
            Mean value of each assets
        cov : ndarray of shape (n_assets, n_assets)
            Covariance matrix
        """
        diag = np.diag(log_cov)
        mu = np.exp(log_mu + np.diag(log_cov) / 2) - 1
        x1 = np.outer(mu, mu)  # Multiply all combinations of the vector mu -> 2-D array
        x2 = np.outer(diag, diag) / 2
        cov = np.exp(x1 + x2) * (np.exp(log_cov) - 1)

        return mu, cov

    def stein_shrinkage(self, cond_cov, shrinkage_factor=(0.2, 0.4)):
        """Stein-type shrinkage of conditional covariance matrices"""
        shrinkage_factor = np.array(shrinkage_factor)

        # Turn it into 3D to make it broadcastable with cond_cov
        shrink_3d = shrinkage_factor[:, np.newaxis, np.newaxis]
        term1 = (1-shrink_3d) * cond_cov

        # Turn term2 into 3D to make it broadcastable with term3
        term2 = (shrinkage_factor * np.trace(cond_cov.T) * 1/self.n_assets)  # Shape (n_states,)
        term3 = np.broadcast_to(np.identity(self.n_assets)[..., np.newaxis],
                                (self.n_assets,self.n_assets,self.n_states)).T  # Shape (n_states, n_assets, n_assets)
        term4 = term2[:, np.newaxis, np.newaxis] * term3
        cond_cov = term1 + term4
        return cond_cov

    def fit_model_get_uncond_dist(self, X, df, n_preds=15, shrinkage_factor=(0.2, 0.4), verbose=False):
        """
        From data, fit hmm model, predict posteriors probabilities and return unconditional distribution.

        Wraps model.fit_predict, get_cond_asset_dist and get_uncond_asset_dist methods into one.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Time series of data
        df : DataFrame of shape (n_samples, n_assets)
            Historical returns for each asset i.
        n_preds : int, default=15
            Number of h predictions
        verbose : boolean, default=False
            Get verbose output

        Returns
        -------
        pred_mu : ndarray of shape (n_preds, n_assets)
            Conditional mean value of each assets
        pred_cov : ndarray of shape (n_preds, n_assets, n_assets)
            Conditional covariance matrix
        """
        self.n_assets = df.shape[1]
        # fit model, return decoded historical state sequnce and n predictions
        # state_sequence is 1D-array with same length as X_rolling
        # posteriors is 2D-array with shape (n_preds, n_states)
        state_sequence, posteriors = self.model.fit_predict(X, n_preds=n_preds, verbose=verbose)

        # Compute conditional mixture distributions in rolling period
        cond_mu, cond_cov = \
            self.get_cond_asset_dist(df, state_sequence)  # shapes (n_states, n_assets), (n_states, n_assets, n_assets)
        cond_cov = self.stein_shrinkage(cond_cov, shrinkage_factor=shrinkage_factor)

        # Transform into unconditional moments at time t
        # Combine with posteriors to also predict moments h steps into future
        # shapes (n_preds, n_assets), (n_preds, n_assets, n_assets)
        pred_mu, pred_cov = self.get_uncond_asset_dist(posteriors, cond_mu, cond_cov)

        return pred_mu, pred_cov, posteriors, state_sequence

class Backtester:
    """
    Backtester for Hidden Markov Models.

    Parameters
    ----------

    Attributes
    ----------
    preds : ndarray of shape (n_samples-window_len, n_preds, n_assets)
        mean predictions for each asset h time steps into the future at each time t.
    cov : ndarray of shape(n_samples-window_len, n_preds, n_assets, n_assets)
        predicted covariance matrix h time steps into the future at each time t.
    """
    def __init__(self, window_len=1700):
        self.preds = None
        self.cov = None
        self.n_states = None
        self.n_assets = None
        self.window_len = window_len

    def rolling_preds_cov_from_hmm(self, X, df_logret, model, n_preds=15, window_len=None, shrinkage_factor=(0.3, 0.3), verbose=False):
        """
        Backtest based on rolling windows.

        Fits a Hidden Markov model within each rolling window and computes the unconditional
        multivariate normal mixture distributions for each asset in the defined universe.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
        Log-returns. Times series data used to train the HMM.
        df_logret : DataFrame of shape (n_samples, n_assets)
            Log-returns. Times series data used when estimating expected returns and covariances.
        model : hidden markov model
            Hidden Markov Model object
        n_preds : int, default=15
            Number of h predictions
        window_len : int, default=1500
        verbose : boolean, default=False
            Make output verbose

        Returns
        -------
        preds : ndarray of shape (n_samples-window_len, n_preds, n_assets)
            Unconditional mean values for each asset
        cov : ndarray of shape (n_samples-window_len, n_preds, n_assets, n_assets)
            Unconditional covariance matrix at each time step t, h steps into future
        """
        self.n_states = model.n_states
        self.n_assets = df_logret.shape[1]

        if window_len == None:  # Ensure class and function window_lens match
            window_len = self.window_len
        else:
            self.window_len = window_len

        finance_hmm = FinanceHMM(model)  # class for computing asset distributions and predictions.

        # Create 3- and 4-D array to store predictions and covariances
        self.preds = np.empty(shape=(len(df_logret) - window_len, n_preds, self.n_assets))  # 3-D array
        self.cov = np.empty(shape=(len(df_logret) - window_len, n_preds, self.n_assets, self.n_assets))  # 4-D array
        self.timestamp = np.empty(shape=len(df_logret) - window_len, dtype=object)

        for t in tqdm.trange(window_len, len(df_logret)):
            # Slice data into rolling sequences
            df_rolling = df_logret.iloc[t-window_len: t]
            X_rolling = X.iloc[t-window_len: t]

            # fit rolling data with model, return predicted means and covariances, posteriors and state sequence
            pred_mu, pred_cov, posteriors, state_sequence = \
                finance_hmm.fit_model_get_uncond_dist(
                    X_rolling, df_rolling, shrinkage_factor=shrinkage_factor, n_preds=n_preds, verbose=verbose)

            self.timestamp[t - window_len] = df_rolling.index[-1]
            self.preds[t - window_len] = pred_mu
            self.cov[t - window_len] = pred_cov

        return self.preds, self.cov

    def backtest_mpc(self, df_rets, preds, covariances, n_preds=15, port_val=1000,
                     start_weights=None, max_drawdown=0.4, max_holding_rf=1.,
                     max_leverage=2.0, gamma_0=5, kappa1=0.008,
                     rho2=0.0005, rho_rf=0.0001, max_holding=0.4, short_cons="LLO",
                     rf_included=True, eps=1e-6):
        """
       Wrapper for backtesting MPC models on given data and predictions.

       Parameters
       ----------
       df_rets : DataFrame of shape (n_samples, n_assets)
           Historical returns for each asset i. Cash must be at the last column position.
       preds : ndarray of shape (n_samples, n_preds, n_assets)
           list of return predictions for each asset h time steps into the future. Each element in list contains,
           from time t, predictions h time steps into the future.
       covariances : ndarray of shape (n_samples, n_preds, n_assets, n_assets)
           list of covariance matrix of returns for each time step t.
       port_val : float, default=1000
           Starting portfolio value.
       start_weights : ndarray of shape (n_assets,)
           Current (known) portfolio weights at the start of backtest. Default is 100% allocation to cash.
           Cash must be the last column in df_rets.
       """
        self.port_val = np.array([0, port_val])
        self.port_ret = np.array([1, 1])
        self.n_assets = df_rets.shape[1]
        self.n_preds = n_preds

        df_rets = df_rets.iloc[-len(preds):]  # Slice returns to match preds

        if start_weights == None:  # Standard init with 100% allocated to cash
            start_weights = np.zeros(self.n_assets)
            start_weights[-1] = 1.
        else:
            start_weights = start_weights

        self.weights = np.zeros(shape=(len(preds) + 1, self.n_assets))  # len(preds) + 1 to include start weights
        self.weights[0] = start_weights

        gamma = np.array([])  # empty array
        trade_cost, turnover = [], []

        # Instantiate MPC object
        mpc_solver = MPC(rets=preds[0], covariances=covariances[0], prev_port_vals=self.port_val,
                         start_weights=self.weights[0], max_drawdown=max_drawdown, gamma_0=gamma_0,
                         kappa1=kappa1, rho2=rho2, rho_rf=rho_rf, max_holding=max_holding, max_holding_rf=max_holding_rf
                         ,max_leverage=max_leverage, short_cons=short_cons, rf_included=rf_included, eps=eps)

        for t in tqdm.trange(preds.shape[0]):
            # Update MPC object
            mpc_solver.rets = np.array(preds[t])
            mpc_solver.cov = np.array(covariances[t])
            mpc_solver.start_weights = self.weights[t]
            mpc_solver.prev_port_vals = self.port_val

            # Solve MPC problem at time t and save weights
            weights_mpc = mpc_solver.cvxpy_solver(verbose=False)  # ndarray of shape (n_preds, n_assets)
            self.weights[t + 1] = weights_mpc[0]  # Only use first forecasted weights
            gamma = np.append(gamma, mpc_solver.gamma)
            delta_weights = self.weights[t] - self.weights[t-1]

            # self.weights and df_rets are one shifted to each other. Time periods should match.
            gross_ret = (self.weights[t + 1] @ (1 + df_rets.iloc[t]))
            shorting_cost = self.short_costs(self.weights[t + 1], rf_return=df_rets.iloc[t, -1])
            trans_cost = self.transaction_costs(delta_weights, trans_cost=0.001)
            port_ret = (gross_ret-shorting_cost) * (1-trans_cost)
            new_port_val = port_ret * self.port_val[-1]
            self.port_ret = np.append(self.port_ret, port_ret)
            self.port_val = np.append(self.port_val, new_port_val)

            trade_cost.append(trans_cost)
            turnover.append(np.linalg.norm(delta_weights, ord=1) / 2)  # Half L1 norm

        self.port_val = self.port_val[1:]  # Throw away first observation since it is artificially set to zero
        self.port_ret = self.port_ret[2:]
        self.gamma = gamma

        # Annualized average trading ost
        self.trans_cost = np.array(trade_cost)
        self.annual_trans_cost = 252 / len(self.trans_cost) * self.trans_cost.sum()

        # Compute average annualized portfolio turnover
        self.daily_turnover = np.array(turnover)
        self.annual_turnover = 252 / len(self.daily_turnover) * self.daily_turnover.sum()

        # Compute return & std.
        n_years = len(self.port_val) / 252
        annual_ret = self.port_ret.prod()**(1/n_years) - 1
        annual_std = self.port_ret.std(ddof=1) * np.sqrt(252)

        return annual_ret, annual_std, self.annual_turnover

    def gridsearch_mpc(self, grid, df_rets, preds, covariances, n_preds=15, port_val=1000,
                     start_weights=None, max_drawdown=1000, max_leverage=2.0, gamma_0=5, kappa1=0.008,
                     rho2=0.0005, max_holding=0.4, short_cons="LO", rf_included=True, eps=1e-6):
        results = pd.DataFrame()
        for max_holding in grid['max_holding']:
            for trans_costs in grid['trans_costs']:
                for holding_costs in grid['holding_costs']:
                    for holding_costs_rf in grid['holding_costs_rf']:
                        print(f"""Computing grid -- max_holding {max_holding} -- trans_costs {trans_costs} holding_costs {holding_costs} holding_costs_rf {holding_costs_rf}""")
                        #try:
                        annual_ret, annual_std, annual_turnover = self.backtest_mpc(
                            df_rets, preds, covariances, n_preds=n_preds, port_val=port_val,
                            start_weights=start_weights, max_drawdown=max_drawdown, max_leverage=max_leverage,
                            gamma_0=gamma_0, kappa1=trans_costs, rho2=holding_costs, rho_rf=holding_costs_rf, max_holding=max_holding,
                            short_cons=short_cons, rf_included=rf_included, eps=eps
                        )

                        results_dict = {'max_holding': max_holding,
                                        'trans_costs': trans_costs,
                                        'holding_costs': holding_costs,
                                        'holding_costs_rf': holding_costs_rf,
                                        'return': annual_ret,
                                        'std': annual_std,
                                        'turnover': annual_turnover}
                        results = results.append(results_dict, ignore_index=True)
                        print(results.tail(1))
                        #except Exception as e:
                        #    print('No convergence')
                        #    print(e)
                        #    continue
        self.gridsearch_df = results
        return results

    def mpc_gammas_shortcons(self, gammas, constraints,
                             data, preds, covariances, n_preds=15, port_val=1000,
                             start_weights=None, max_holding_rf=1.,
                             max_leverage=2.0, trans_costs=0.001,
                             holding_costs=0.0000, max_holding=0.2, eps=1e-6):

        df = pd.DataFrame()


        for constr in constraints:
            print(f'Backtesting for params {constr}')
            results = {f'gamma_{i}': [] for i in gammas}
            short_con = constr[0]
            max_drawdown = constr[1]
            for gamma in gammas:
                self.backtest_mpc(data.rets, preds, covariances, n_preds=n_preds, port_val=port_val,
                                  start_weights=start_weights, max_drawdown=max_drawdown, max_leverage=max_leverage,
                                  gamma_0=gamma, kappa1=trans_costs, rho2=holding_costs, max_holding=max_holding,
                                  short_cons=short_con, eps=eps)

                results[f'gamma_{gamma}'] = self.port_val

            df_temp = pd.DataFrame(results)
            df_temp['short_cons'] = short_con
            df_temp['D_max'] = max_drawdown
            df_temp['timestamp'] = data.rets.index[-len(df_temp):]
            df_temp['T-bills rf'] = data.prices['T-bills rf'].iloc[-len(df_temp):].values
            df = df.append(df_temp)

        # self.annual_turnover, self.annual_trans_cost, self.port_val
        self.port_val_df = df
        return df

    def mpc_shortcons(self, constraints,
                     data, preds, covariances, n_preds=15, port_val=1000,
                     start_weights=None, max_holding_rf=1.,
                     max_leverage=2.0, trans_costs=0.001,
                     holding_costs=0.0000, max_holding=0.2, eps=1e-6):

        df = pd.DataFrame()
        results = {f'{constr[0]}_{constr[1]}': [] for constr in constraints}
        for constr in constraints:
            print(f'Backtesting for params {constr}')
            short_con = constr[0]
            max_drawdown = constr[1]

            self.backtest_mpc(data.rets, preds, covariances, n_preds=n_preds, port_val=port_val,
                              start_weights=start_weights, max_drawdown=max_drawdown, max_leverage=max_leverage,
                              gamma_0=5, kappa1=trans_costs, rho2=holding_costs, max_holding=max_holding,
                              short_cons=short_con, eps=eps)

            results[f'{constr[0]}_{constr[1]}'] = self.port_val

        df = pd.DataFrame(results)
        df['timestamp'] = data.rets.index[-len(df):]
        df['T-bills rf'] = data.prices['T-bills rf'].iloc[-len(df):].values

        # self.annual_turnover, self.annual_trans_cost, self.port_val
        self.port_val_df = df
        return df

    def backtest_equal_weighted(self, df_rets, rebal_freq='M', port_val=1000, use_weights=None, start_weights=None):
        """
       Backtest an equally weighted portfolio, with specified rebalancing frequency.

       Parameters
       ----------
       df_rets : DataFrame of shape (n_samples, n_assets)
           Historical returns for each asset i. Cash must be at the last column position.
       rebal_freq : int, default=20
            Rebalance frequency. Default is 20, i.e monthly.
       port_val : float, default=1000
           Starting portfolio value.
       start_weights : ndarray of shape (n_assets,)
           Current (known) portfolio weights at the start of backtest. Default is 100% allocation to cash.
           Cash must be the last column in df_rets.
       """
        self.port_val = np.array([0, port_val])
        self.n_assets = df_rets.shape[1]

        if np.any(use_weights) == None:
            use_weights = np.array([1 / self.n_assets] * self.n_assets)  # Vector of shape (n_assets,)


        if start_weights == None:  # Standard init with 100% allocated to cash
            start_weights = np.zeros(self.n_assets)
            start_weights[-1] = 1.
        else:
            start_weights = start_weights

        weights = start_weights
        trade_cost, turnover = [], []

        # Group data into months - average sample size is 20
        # Then for each month loop over the daily returns and update weights
        # The problem is recursive and thus requires looping done this way
        for month_dt, df_group in tqdm.tqdm(df_rets.groupby(pd.Grouper(freq=rebal_freq))):
            # Compute transaction costs for each month. Subtracted from gross ret the first of the month
            delta_weights = use_weights - weights
            trans_cost = self.transaction_costs(delta_weights)
            weights = use_weights  # Reset weights
            for day in range(len(df_group)):
                # Calculate gross returns for portfolio and append it
                if day == 0:
                    gross_ret = (1 + df_group.iloc[day]) * (1-trans_cost)
                else:
                    gross_ret = 1 + df_group.iloc[day]

                new_port_val = weights @ gross_ret * self.port_val[-1]
                self.port_val = np.append(self.port_val, new_port_val)

                new_w = gross_ret * weights
                new_w /= new_w.sum()  # Weights sum to 1
                weights = new_w  # Update weights each iteration

            trade_cost.append(trans_cost)
            turnover.append(np.linalg.norm(delta_weights, ord=1) / 2)  # Half L1 norm

        self.port_val = self.port_val[1:]  # Throw away first observation since it is artificially set to zero

        # Annualized average trading ost
        self.trans_cost = np.array(trade_cost)
        self.annual_trans_cost = 12 / len(self.trans_cost) * self.trans_cost.sum()

        # Compute average annualized portfolio turnover
        self.monthly_turnover = np.array(turnover)
        self.annual_turnover = 12 / len(self.monthly_turnover) * self.monthly_turnover.sum()

    def short_costs(self, weights, rf_return):
        """
        Compute shorting costs, assuming a fee equal to the risk-free asset is paid.
        """
        weights_no_rf = weights[:-1]  # Remove risk-free asset from array
        short_weights = weights_no_rf[weights_no_rf < 0.0].sum()  # Sum of all port weights below 0.0
        return -short_weights * rf_return

    def transaction_costs(self, delta_weights, trans_cost=0.001):
        """
        Compute transaction costs. Assumes no costs in risk-free asset and equal cost to
        buying and selling assets.
        """
        delta_weights = delta_weights[:-1]  # Remove risk-free asset as it doesn't have trading costs
        delta_weights = np.abs(delta_weights).sum()  # abs since same price for buying/selling

        return delta_weights * trans_cost

    def asset_metrics(self, df_prices):
        """Compute performance metrics for a given portfolio/asset"""
        df_ret = df_prices.pct_change().dropna()
        n_years = len(df_ret) / 252

        # Get regular cagr and std
        ret = df_ret.drop('T-bills rf', axis=1)
        cagr = ((1 + ret).prod(axis=0)) ** (1 / n_years) - 1
        std = ret.std(axis=0, ddof=1) * np.sqrt(252)

        # Compute metrics in excess of the risk-free asset
        excess_ret = df_ret.subtract(df_ret['T-bills rf'], axis=0).drop('T-bills rf', axis=1)
        excess_cagr = ((1 + excess_ret).prod(axis=0)) ** (1 / n_years) - 1
        excess_std = excess_ret.std(axis=0 ,ddof=1) * np.sqrt(252)
        sharpe = excess_cagr / excess_std

        df_prices = df_prices.drop('T-bills rf', axis=1)
        peaks = df_prices.cummax(axis=0)
        drawdown = -(df_prices - peaks) / peaks
        max_drawdown = drawdown.max(axis=0)
        calmar = excess_cagr / max_drawdown

        metrics = {'return': cagr,
                   'std': std,
                   'excess_return': excess_cagr,
                   'excess_std': excess_std,
                   'sharpe': sharpe,
                   'max_drawdown': max_drawdown,
                   'calmar_ratio': calmar}

        metrics = pd.DataFrame(metrics)

        return metrics

    def single_port_metric(self, df_prices, port_val, compare_assets=False):
        """Compute performance metrics for a given portfolio/asset"""
        # Merge port_val with data
        df_prices = df_prices.iloc[-len(port_val):]
        df_prices['port_val'] = port_val
        df_prices.dropna(inplace=True)
        df_ret = df_prices.pct_change().dropna()

        # Annual returns, std
        n_years = len(port_val) / 252
        excess_ret = df_ret['port_val'] - df_ret['T-bills rf']

        excess_cagr = ((1+excess_ret).prod())**(1/n_years) - 1
        excess_std = excess_ret.std(ddof=1) * np.sqrt(252)
        sharpe = excess_cagr / excess_std

        # Drawdown
        peaks = np.maximum.accumulate(port_val)
        drawdown = -(port_val-peaks) / peaks
        max_drawdown = np.max(drawdown)
        max_drawdown_end = np.argmax(drawdown)
        max_drawdown_beg = np.argmax(port_val[:max_drawdown_end])
        drawdown_dur = max_drawdown_end - max_drawdown_beg  # TODO not showing correct values
        calmar = excess_cagr / max_drawdown

        metrics = {'excess_return': excess_cagr,
                   'excess_std': excess_std,
                   'sharpe': sharpe,
                   'max_drawdown': max_drawdown,
                   'max_drawdown_dur': drawdown_dur,
                   'calmar_ratio': calmar}

        return metrics

    def mulitple_port_metrics(self, df_port_val):
        """Compute performance metrics for a given portfolio/asset"""
        df = pd.DataFrame()
        for type, df_groupby in df_port_val.groupby(['short_cons', 'D_max']):
            df_prices = df_groupby.drop(columns=['short_cons', 'D_max', 'timestamp'])
            df_rets = df_prices.pct_change().dropna()

            # Annual returns, std
            n_years = len(df_rets) / 252
            ret = df_rets.drop('T-bills rf', axis=1)
            cagr = ((1 + ret).prod(axis=0)) ** (1 / n_years) - 1
            std = ret.std(axis=0, ddof=1) * np.sqrt(252)

            excess_ret = df_rets.subtract(df_rets['T-bills rf'], axis=0).drop('T-bills rf', axis=1)
            excess_cagr = ((1 + excess_ret).prod(axis=0)) ** (1 / n_years) - 1
            excess_std = excess_ret.std(axis=0 ,ddof=1) * np.sqrt(252)
            sharpe = excess_cagr / excess_std

            df_prices = df_prices.drop('T-bills rf', axis=1)
            peaks = df_prices.cummax(axis=0)
            drawdown = -(df_prices - peaks) / peaks
            max_drawdown = drawdown.max(axis=0)
            """
            max_drawdown_end = np.argmax(drawdown, axis=0)
            max_drawdown_beg = np.argmax(drawdown[:max_drawdown_end], axis=0)
            drawdown_dur = max_drawdown_end - max_drawdown_beg  # TODO not showing correct values
            """
            calmar = excess_cagr / max_drawdown

            metrics = {'return': cagr,
                       'std': std,
                       'excess_return': excess_cagr,
                       'excess_std': excess_std,
                       'sharpe': sharpe,
                       'max_drawdown': max_drawdown,
                       'calmar_ratio': calmar}

            df_temp = pd.DataFrame(metrics)
            df_temp['short_cons'] = type[0]
            df_temp['D_max'] = type[1]
            df = df.append(df_temp)

        return df

    def plot_port_val(self, data, mpc_val, equal_w_val, start=None, savefig=None):
        # Prepare data
        equal_w_val = equal_w_val[-len(mpc_val):]
        data.dropna(inplace=True)
        data = data.iloc[-len(mpc_val):]
        data['MPC'] = mpc_val
        data['1/n'] = equal_w_val
        data = data[['MPC', '1/n']]  # Drop all other cols

        if not start == None:
            data = data.loc[start:]

        data = data / data.iloc[0] * 100

        # Plotting
        plt.rcParams.update({'font.size': 15})
        fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True, figsize=(15,10))

        ax.plot(data.index, data)
        # ax[0].set_yscale('log')
        ax.set_ylabel('$P_t$')

        plt.tight_layout()

        if not savefig == None:
            plt.savefig('./images/' + savefig)
        plt.show()

if __name__ == "__main__":
    data = DataPrep(out_of_sample=True)
    df_port_val = data.prices
    df_ret = data.rets
    df_logret = data.logrets
    X = df_logret["S&P 500 "]

    model1 = MLEHMM(n_states=2, init="random", random_state=42, epochs=20, max_iter=100)
    backtester = Backtester()

    #preds = np.load(path + 'rolling_preds.npy')
    #cov = np.load(path + 'rolling_cov.npy')
    #weights, port_val, gamma = backtester.backtest_mpc(df_ret, preds, cov, short_cons='LLO')
    #port_val = np.load(path + 'port_val.npy')
    #weights = np.load(path + 'mpc_weights.npy')
    #metrics = backtester.asset_metrics(df_prices)




import numpy as np
import pandas as pd
import tqdm
from numpy import ndarray
from sklearn.cluster._kmeans import kmeans_plusplus
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import StandardScaler

from hmmpy.sampler import SampleHMM
from hmmpy.base import BaseHiddenMarkov
from hmmpy import hmm_cython


#pyximport.install()  # TODO can only be active during development -- must be done through setup.py

class JumpHMM(BaseHiddenMarkov):

    def __init__(self, n_states: int = 2, jump_penalty: float = 16, window_len: tuple = (6, 14),
                 init: str = 'kmeans++', max_iter: int = 30, tol: int = 1e-6,
                 epochs: int = 10, random_state: int = 42):
        super().__init__(n_states, init, max_iter, tol, epochs, random_state)

        # Init parameters initial distribution, transition matrix and state-dependent distributions
        self.type = 'jump'
        self.jump_penalty = jump_penalty
        self.window_len = window_len
        self.theta = None
        self.state_seq = None
        self.n_features = None

    def _init_params(self, Z, X=None, diag_uniform_dist=(.95, .99), output_hmm_params=False):
        super()._init_params()

        if self.init == "kmeans++":
            if not isinstance(Z, np.ndarray):
                raise Exception("To initialize with kmeans++ a sequence of data must be provided in an ndarray")
            if Z.ndim == 1:
                Z = Z.reshape(-1, 1)  # Compatible with sklearn

            # Theta
            centroids, _ = kmeans_plusplus(Z, self.n_states)  # Use sklearns kmeans++ algorithm
            theta = centroids.T  # Transpose to get shape (n_features, n_states)

            # State sequence
            state_seq, _ = self._fit_state_seq(Z, theta)

            if output_hmm_params == True:  # output all hmm parameters from state sequence
                if not X == None:
                    self.get_params_from_seq(X, state_sequence=self.state_seq)
                else:
                    raise Exception('''To output hmm-params from sequence, X must be provided as a 1D-array.
                                    X has to be non-standardized data otherwise it will return standardized results''')

            return state_seq, theta

    def construct_features(self, X: ndarray, window_len: tuple, feature_set='feature_set_3'):
        if not feature_set in ['feature_set_1', 'feature_set_2', 'feature_set_3']:
            raise Exception('Invalid feature set. Valid options are [feature_set_1, feature_set_2, feature_set_3]')

        df = pd.DataFrame({'raw_input': X})

        if feature_set == 'feature_set_1':
            # Absolute changes
            df['left_abs_change'] = np.abs(df['raw_input'].diff())  # np.abs(np.diff(X))
            df['prev_left_abs_change'] = df['left_abs_change'].shift(1)

            # Rolling features
            for i in range(len(window_len)):
                rolling_window = df['raw_input'].rolling(window_len[i])

                df['left_local_mean_'+str(i)] = rolling_window.mean()
                df['left_local_std_'+str(i)] = rolling_window.std(ddof=1)
                df['left_local_median_'+str(i)] = rolling_window.median() #
                df['min_max_diff_'+str(i)] = rolling_window.max() - rolling_window.min()

        elif feature_set == 'feature_set_2':
            for i in range(len(window_len)):
                rolling_window = df['raw_input'].rolling(window_len[i])
                #Bollinger bands (Remember middle band is equal to the rolling mean)
                df['upper_bollinger_band'] = rolling_window.mean() + (1.96 * rolling_window.std(ddof=1))
                df['lower_bollinger_band'] = rolling_window.mean() - (1.96 * rolling_window.std(ddof=1))
                df['band_width'] = df['upper_bollinger_band'] - df['lower_bollinger_band']
                df['bbands_%_bi'] = (rolling_window.mean() - df['lower_bollinger_band']) / (df['upper_bollinger_band'] - df['lower_bollinger_band']) # %B quantifies a security's average price relative to the upper and lower Bollinger Band.

            #Exponential moving average 10 days
            df['ema_10'] = df['raw_input'].ewm(span=10).mean()

            df.drop('raw_input', inplace=True, axis=1) #Active this for 2nd feature set!

        elif feature_set == 'feature_set_3':
            df['left_abs_change'] = np.abs(df['raw_input'].diff())  # np.abs(np.diff(X))
            df['prev_left_abs_change'] = df['left_abs_change'].shift(1)

            # Rolling features
            for i in range(len(window_len)):
                rolling_window = df['raw_input'].rolling(window_len[i])

                df['left_local_mean_' + str(i)] = rolling_window.mean()
                df['left_local_std_' + str(i)] = rolling_window.std(ddof=1)
                df['left_local_median_' + str(i)] = rolling_window.median()  #
                df['min_max_diff_' + str(i)] = rolling_window.max() - rolling_window.min()

                # Bollinger bands (Remember middle band is equal to the rolling mean)
                df['upper_bollinger_band'] = rolling_window.mean() + (1.96 * rolling_window.std(ddof=1))
                df['lower_bollinger_band'] = rolling_window.mean() - (1.96 * rolling_window.std(ddof=1))
                df['band_width'] = df['upper_bollinger_band'] - df['lower_bollinger_band']
                df['bbands_%_bi'] = (rolling_window.mean() - df['lower_bollinger_band']) / (
                df['upper_bollinger_band'] - df['lower_bollinger_band'])  # %B quantifies a security's average price relative to the upper and lower Bollinger Band.

            # Exponential moving average
            df['ema_10'] = df['raw_input'].ewm(span=10).mean()



        Z = df.dropna().to_numpy()
        # Scale features
        scaler = StandardScaler()
        Z = scaler.fit_transform(Z)

        #Consider defining feature set from PCA?

        #pca = PCA(n_components=2)
        #pca_features = pca.fit_transform(Z)
        #Z['PCA_1'] = pca_features[:,0]
        #Z['PCA_2'] = pca_features[:,1]
        #Z = Z.loc[:,'PCA_1':'PCA_2']
        self.n_features = Z.shape[1]

        return Z

    def _l2_norm_squared(self, z, theta):
        """
        Compute the squared l2 norm at each time step.

        Squared l2 norm is computed as ||z_t - theta ||_2^2.

        Parameters
        ----------
        z : ndarray of shape (n_samples, n_features)
            Data to be fitted
        theta : ndarray of shape (n_features, n_states)
            jump model parameters
        Returns
        -------
        norms: ndarray of shape (n_samples, n_states)
            Squared l2 norms conditional on latent states
        """
        norms = np.zeros(shape=(len(z), self.n_states))

        for j in range(self.n_states):
            diff = theta[:, j] - z  # ndarray of shape (n_samples, n_states) with differences
            norms[:, j] = np.square(np.linalg.norm(diff, axis=1))  # squared state conditional l2 norms

        return norms  # squared l2 norm.

    def _fit_theta(self, Z, state_seq):
        """
        Fit theta, i.e minimize the squared L2 norm in each latent state.
        Computed analytically. See notebooks/math_overviews_algos for proof of solution.

        Parameters
        ----------
        Z : ndarray of shape (n_samples, n_features)
            Set of standardized times series features
        state_seq : ndarray of shape (n_samples,)
            State sequence

        Returns
        -------
        theta : ndarray of shape (n_features, n_states)
            jump model parameters. Distances from state (cluster) centers.
        """
        theta = np.zeros(shape=(self.n_features, self.n_states))
        for j in range(self.n_states):
            state_slicer = np.array(state_seq == j)  # Boolean array of True/False
            N_j = np.sum(state_slicer)  # Number of terms in state j

            if N_j != 0:
                z = Z[state_slicer].sum(axis=0)  # Sum each feature across time
                theta[:, j] = z / N_j
            else:
                theta[:, j] = 0.  # TODO what is the best estimate of theta when the state is not present?

        return theta

    def _fit_state_seq(self, X, theta):
        """
        Fit a state sequence based on current theta estimates.
        Used in jump model fitting. Uses a dynamic programming technique very
        similar to the viterbi algorithm.

        Parameters
        ----------
        X : ndarray of shape (n_samples)
            Standardized returns.
        theta : ndarray of shape (n_features, n_states)
            jump model parameters. Distances from state (cluster) centers.

        Returns
        -------
        state_seq : ndarray of shape (n_samples,)
            State sequence
        objective_score : float
            Objective score under the model
        """
        l2_norms = self._l2_norm_squared(X, theta)  # Compute array of all squared l2 norms
        n_samples, _ = l2_norms.shape

        losses, state_preds = hmm_cython.jump_state_seq(n_samples, self.n_states,
                                                        self.n_features,
                                                        self.jump_penalty,
                                                        l2_norms)
        # Compute score of objective function
        all_likelihoods = losses[np.arange(len(losses)), state_preds].sum()
        state_changes = np.diff(state_preds) != 0  # True/False array showing state changes
        jump_penalty = (state_changes * self.jump_penalty).sum()  # Multiply all True values with penalty

        objective_score = all_likelihoods + jump_penalty  # Float

        return state_preds, objective_score

    def _fit(self, X, Z, verbose=False):
        """ Container for the loop where fitting happens"""
        best_objective_score = np.inf

        # Each epoch independently inits and fits a new model to the same data
        for epoch in range(self.epochs):  # TODO consider parallel processing
            # Do new init at each epoch
            state_seq, theta = self._init_params(Z, output_hmm_params=False, X=X)
            old_state_seq = state_seq
            old_objective_score = np.inf

            for iter in range(self.max_iter):
                theta = self._fit_theta(Z, state_seq)
                state_seq, objective_score = self._fit_state_seq(Z, theta)

                # Check convergence criterion
                crit1 = np.array_equal(state_seq, old_state_seq)  # No change in state sequence
                crit2 = np.abs(old_objective_score - objective_score)
                if crit1 is True or crit2 < self.tol:
                    self.is_fitted = True
                    # If model is converged check if current epoch is the best
                    # If current model is best all model params are updated
                    if objective_score < best_objective_score:
                        best_objective_score = objective_score
                        self.theta = theta
                        self.state_seq = state_seq

                        if verbose == 2:
                            print(
                                f'Epoch {epoch} -- Iter {iter} -- likelihood {objective_score} -- Theta {self.theta[0]} ')

                    break  # Break out of inner loop and go to next epoch

                else:
                    old_state_seq = state_seq
                    old_objective_score = objective_score

    def fit(self, X, get_hmm_params=True, sort_state_seq=True, verbose=False, feature_set='feature_set_3'):
        self.is_fitted = False
        Z = self.construct_features(X, window_len=self.window_len, feature_set=feature_set)

        self._fit(X=X, Z=Z, verbose=verbose)  # Container for loop where fitting happens

        if self.is_fitted is False:
            max_iter = self.max_iter
            self.max_iter = max_iter * 2  # Double amount of iterations
            self._fit(X, Z)  # Try fitting again
            self.max_iter = max_iter  # Reset max_iter back to user-input
            if self.is_fitted == False and verbose == True:
                print(
                    f'Jump NOT FITTED -- epochs {self.epochs} -- iters {self.max_iter * 2}')

        if get_hmm_params is True and self.is_fitted is True:
            if sort_state_seq is True:
                self.state_seq = self._check_state_sort(X, self.state_seq)
            self.get_params_from_seq(X, state_sequence=self.state_seq)

    def get_params_from_seq(self, X, state_sequence):
        """
        Stores and outputs the model parameters based on the input sequence.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Time series of data

        state_sequence : ndarray of shape (n_samples)
            State sequence for a given observation sequence
        Returns
        ----------
        Hmm parameters
        """
        # Slice data
        if X.ndim == 1:  # Makes function compatible on higher dimensions
            X = X[(self.window_len[-1] - 1):]
        else:
            raise Exception('''X must be one-dimensional. Pass raw data as 1-D array.''')

        # group by states
        diff = np.diff(state_sequence)
        df_states = pd.DataFrame({'state_seq': state_sequence,
                                  'X': X,
                                  'state_sojourns': np.append([False], diff == 0),
                                  'state_changes': np.append([False], diff != 0)})

        state_groupby = df_states.groupby('state_seq')

        # Conditional distributions
        self.mu = state_groupby['X'].mean().values.T  # transform mean back into 1darray
        self.std = state_groupby['X'].std(ddof=1).values.T

        # Transition probabilities
        # TODO only works for a 2-state HMM
        tpm = np.diag(state_groupby['state_sojourns'].sum())
        state_changes = state_groupby['state_changes'].sum()

        # if only 1 state is detected make the other state zero
        if state_changes.sum() > 0:
            tpm[0, 1] = state_changes[0]
            tpm[1, 0] = state_changes[1]

        else:
            tpm = np.array([[1., 0.], [1., 0.]])
            self.mu = np.append(self.mu, 0.)
            self.std = np.append(self.std, 0.)

        self.tpm = tpm / tpm.sum(axis=1).reshape(-1, 1)  # make rows sum to 1

        # init dist and stationary dist
        self.start_proba = np.zeros(self.n_states)
        self.start_proba[state_sequence[0]] = 1.
        self.stationary_dist = self.get_stationary_dist(tpm=self.tpm)

    def _check_state_sort(self, X, state_sequence):  # TODO give this same structure as in EM_HMM
        """
        Checks whether the low-variance state is the first state.

        Otherwise sorts state predictions accoridngly.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Data to be fitted.
        state_sequence : ndarray of shape (n_samples,)
            Predicted state sequence.

        Returns
        -------
        state_sequence : ndarray of shape (n_samples,)
            Predicted state sequence sorted in correct order.
        """
        # Slice data
        if X.ndim == 1:  # Makes function compatible on higher dimensions
            X = X[(self.window_len[-1] - 1):]
        else:
            raise Exception('''X must be one-dimensional. Pass raw data as 1-D array.''')

        df_states = pd.DataFrame({'state_seq': state_sequence,
                                  'X': X})

        state_groupby = df_states.groupby('state_seq')
        self.std = state_groupby['X'].std(ddof=1).values.T

        # Sort array ascending and check if order is changed
        # If the order is changed then states are reversed
        if np.sort(self.std)[0] != self.std[0]:
            state_sequence = np.where(state_sequence == 0, 1, 0)

        return state_sequence

    def bac_score_1d(self, X, y_true, jump_penalty, window_len=(6,14), verbose=False):
        self.jump_penalty = jump_penalty
        self.fit(X, get_hmm_params=False)  # Updates self.state_seq
        self.state_seq = self._check_state_sort(X, self.state_seq)

        y_pred = self.state_seq
        y_true = y_true[(self.window_len[-1] - 1):]  # slice y_true to have same dim as y_pred

        conf_matrix = confusion_matrix(y_true, y_pred)
        keep_idx = conf_matrix.sum(axis=1) != 0
        conf_matrix = conf_matrix[keep_idx]

        tp = np.diag(conf_matrix)
        fn = conf_matrix.sum(axis=1) - tp
        tpr = tp / (tp + fn)
        bac = np.mean(tpr)

        if verbose is True:
            logical_1 = bac < 0.5
            logical_2 = conf_matrix.ndim > 1 and \
                        (np.any(conf_matrix.sum(axis=1)==0) and \
                        jump_penalty < 100)

            if logical_1 or logical_2:
                print(f'bac {bac} -- tpr {tpr} -- jump_penalty {jump_penalty}')
                print(conf_matrix)

        return bac

    def bac_score_nd(self, X, y_true, jump_penalty, window_len=(6, 14)):
        bac = []
        for seq in tqdm.tqdm(range(X.shape[1])):
            bac_temp = self.bac_score_1d(X[:, seq], y_true[:, seq], jump_penalty, window_len=self.window_len)
            bac.append(bac_temp)

        return bac


if __name__ == '__main__':
    model = JumpHMM(n_states=2, jump_penalty=0.0025, window_len=(6, 14), random_state=2)
    sampler = SampleHMM(n_states=2, random_state=2)

    n_samples = 1000
    n_sequences = 1
    X, viterbi_states, true_states = sampler.sample_with_viterbi(n_samples, n_sequences)


    #plotting.plot_samples_states_viterbi(X[:,0], viterbi_states[:,0], true_states[:,0])

    #for i in range(50):
       # bac = model.bac_score_1d(X[:,i], viterbi_states[:, i] , 30)

    model.fit(X, verbose=True)







from setuptools import Extension, setup, find_packages
from setuptools.command.build_ext import build_ext


#from distutils.extension import Extension
#from Cython.Distutils import build_ext

#ext_modules = [Extension("hmm_cython", ["hidden_markov/hmm_cython.pyx"])]

class build_ext(build_ext):
    def finalize_options(self):
        from Cython.Build import cythonize
        import numpy as np
        import numpy.distutils

        self.distribution.ext_modules[:] = cythonize("hmmpy/*.pyx", language_level='3')
        # Sadly, this part needs to be done manually.
        for ext in self.distribution.ext_modules:
            for k, v in np.distutils.misc_util.get_info("npymath").items():
                setattr(ext, k, v)
            ext.include_dirs = [np.get_include()]

        super().finalize_options()

    def build_extensions(self):
        try:
            self.compiler.compiler_so.remove("-Wstrict-prototypes")
        except (AttributeError, ValueError):
            pass
        super().build_extensions()

setup(
    name='hmm-py',
    version='0.0.0',
    url='https://github.com/Cstolborg/hmmpy',
    author='Cstolborg',
    author_email='christianstolborg@gmail.com',
    description='A python module to implement Hidden Markov hidden_markov for financial times series.',
    long_description=open("README.md", 'r').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers =[
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.8',
    setup_requires=[
        'Cython',
        'numpy>=1.20.1'
    ],
    install_requires=[
        "numpy>=1.20.1",
        "scikit-learn>=0.24.0",
        "scipy>=1.5.4",
        "pandas>=1.2.0",
        "cvxpy==1.1.10",
        "tqdm",
        'cython',
    ],
    extras_require={
        'docs': ['Sphinx', 'sphinx-gallery']
    },

    cmdclass={'build_ext': build_ext},
    ext_modules=[Extension("", [])]
    #ext_modules=cythonize('hidden_markov/hmm_cython.pyx', include_path=[np.get_include()])
    #cmdclass = {'build_ext': build_ext},
    #ext_modules=cythonize('hidden_markov/hmm_cython.pyx')
    #include_dirs = [numpy.get_include()]
)




""" TODO
Daily data begins:
    MSCI Emerging: 1987-12-31
    Barclays : 1994-02-28
    T-bills rf : 1997-05-19
    DAX : 1999-01-04
    Real Estate: 1999-01-04
    Hedge funds : 2003-03-31
    PE : 2003-03-31

"""
import datetime as dt
from dataclasses import dataclass, field
from typing import List, Dict

import numpy as np
import pandas as pd
from scipy import stats

pd.set_option('display.max_columns', 10); pd.set_option('display.width', 320)
np.seterr(divide='ignore')

@dataclass
class DataPrep:
    """ Class used for loading and preparing data for analysis.

    Primary use case is getting prices, returns and logreturns.
    """
    path: str = '../../data/price_series.csv'
    out_of_sample: bool = True
    drop_insample: List = field(default_factory= lambda: ['Hedge Funds Global', 'Private Equity', 'DAX ',
                  'MSCI World', 'European Public Real Estate'])
    drop_outsample: List = field(default_factory= lambda: ['MSCI World', 'DAX '])
    insample_start: str = '1994-02-28'
    insample_end: str = '2003-11-20'

    def __post_init__(self):
        self.prices = self.load_prices()
        self.rets = self.load_returns()
        self.logrets = self.load_logreturns()
        self.long_logrets = self.load_long_series_logret(outlier_corrected=False)

    def load_prices(self):
        df_prices = pd.read_csv(self.path, index_col='Time', parse_dates=True)

        if self.out_of_sample is True:
            df_prices = df_prices.drop(self.drop_outsample, axis=1)
        else:
            # In-sample data
            df_prices = df_prices.drop(self.drop_insample, axis=1)
            df_prices = df_prices.loc[self.insample_start:self.insample_end]  # Daily data Barclays from 1994 and oos begins 2003

        df_prices.interpolate(method='linear', inplace=True)
        df_prices.dropna(inplace=True)

        return df_prices

    def load_returns(self):
        df_ret = self.prices.pct_change()
        df_ret.dropna(inplace=True)

        return df_ret

    def load_logreturns(self):
        df_ret = np.log(self.prices) - np.log(self.prices.shift(1))
        df_ret.dropna(inplace=True)

        return df_ret

    def load_long_series_logret(self, outlier_corrected=False, threshold=4):
        df = pd.read_csv(self.path, index_col='Time', parse_dates=True)
        df = df['S&P 500 ']
        df_ret = np.log(df) - np.log(df.shift(1))
        df_ret.dropna(inplace=True)

        # Remove trading days with movements above 10%.
        # This includes black monday, two days during GFC, and one day during COVID
        mean, std = df_ret.mean(), df_ret.std()
        df_ret.loc[df_ret >= 0.1] = mean + 6*std
        df_ret.loc[df_ret <= -0.1] = mean - 6*std

        # Remove all observations with std's above 4
        if outlier_corrected is True:
            df_ret = self.replace_outliers(df_ret, threshold=threshold)

        return df_ret

    @staticmethod
    def replace_outliers(data, threshold=4):
        """ Replaces outliers further than x STDs away from mean by the threshold value """
        zscores = stats.zscore(data)
        mean, std = data.mean(), data.std()
        data.loc[zscores >= threshold] = mean + std * threshold
        data.loc[zscores <= -threshold] = mean - std * threshold

        return data

    @staticmethod
    def moving_average(a, n=10):
        a = np.array(a)
        ret = np.cumsum(a)
        ret[n:] = ret[n:] - ret[:-n]
        return ret[n - 1:] / n


######## ALL BELOW FUNCTIONS ARE HERE FOR BACKWARD COMPATIBILITY ONLY. TO BE DEPRECATED #######################

def load_prices(path='../../data/price_series.csv', out_of_sample=True):
    df = pd.read_csv(path, index_col='Time', parse_dates=True)

    if out_of_sample is True:
        df = df.drop(['MSCI World', 'DAX '], axis=1)
    else:
        # In-sample data
        df = df.drop(['Hedge Funds Global', 'Private Equity', 'DAX ',
                  'MSCI World', 'European Public Real Estate'] ,axis=1)
        df = df.loc['1994-02-28':'2003-11-20']  # Daily data Barclays from 1994 and oos begins 2003

    df.interpolate(method='linear', inplace=True)
    df.dropna(inplace=True)

    return df

def load_returns(path='../../data/price_series.csv', out_of_sample=True):
    df = load_prices(path ,out_of_sample=out_of_sample)
    df_ret = df.pct_change()
    df_ret.dropna(inplace=True)

    return df_ret

def load_logreturns(path='../../data/price_series.csv', out_of_sample=True):
    df = load_prices(path, out_of_sample=out_of_sample)
    df_ret = np.log(df) - np.log(df.shift(1))
    df_ret.dropna(inplace=True)
    df_ret = df_ret.loc['1994-02-28':]  # We start here because otherwise too many assets have monthly values only.

    return df_ret

def load_long_series_logret(path='../../data/price_series.csv', outlier_corrected=False):
    df = pd.read_csv(path, index_col='Time', parse_dates=True)
    df = df['S&P 500 ']
    df_ret = np.log(df) - np.log(df.shift(1))
    df_ret.dropna(inplace=True)

    # Remove all observations with std's above 4
    if outlier_corrected is True:
        df_ret = df_ret[(np.abs(stats.zscore(df_ret)) < 4)]

    return df_ret

def moving_average(a, n=10) :
    a = np.array(a)
    ret = np.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def get_cov_mat(df_ret):
    return df_ret.cov()


if __name__ == '__main__':
    path = '../../data/price_series.csv'
    data = DataPrep(path=path, out_of_sample=True)
    print(data.prices)
    print(data.rets)
    print(data.logrets)

    #data.load_prices()


import sys

import numpy as np
from scipy import stats
from scipy.special import logsumexp
from sklearn.base import BaseEstimator
from sklearn.metrics import confusion_matrix

from hmmpy import hmm_cython


class BaseHiddenMarkov(BaseEstimator):
    """
    Parent class for Hidden Markov methods with gaussian distributions.
    Contain methods related to:
    1. Initializing HMM parameters
    2. Methods that assumes an HMM is fitted and are used for sampling, prediction etc.

    To fit HMMs refer to the respective child classes

    Parameters
    ----------
    n_states : int, default=2
        Number of hidden states
    max_iter : int, default=100
        Maximum number of iterations to perform during expectation-maximization
    tol : float, default=1e-6
        Criterion for early stopping
    epochs : int, default=1
        Number of complete passes through the data to improve fit
    random_state : int, default = 42
        Parameter set to recreate output
    init : str
        Set to 'kmeans++' to use that init method - only supported for jump hidden_markov.
        Set to 'random' for random initialization.
        Set to "deterministic" for deterministic init.

    Attributes
    ----------
    mu : ndarray of shape (n_states,)
        Fitted means for each state
    std : ndarray of shape (n_states,)
        Fitted std for each state
    tpm : ndarray of shape (n_states, n_states)
        Transition probability matrix between states
    start_proba : ndarray of shape (n_states,)
        Initial state occupation distribution
    """

    def __init__(self, n_states: int = 2, init: str = 'random', max_iter: int = 100, tol: float = 1e-6,
                 epochs: int = 1, random_state: int = 42):
        self.n_states = n_states
        self.random_state = random_state
        self.max_iter = max_iter  # Max iterations to fit model
        self.epochs = epochs  # Set number of random inits used in model fitting
        self.tol = tol
        self.init = init

        self.stationary_dist = None
        self.tpm = None
        self.start_proba = None
        self.mu = None
        self.std = None
        self.is_fitted = False

        np.random.seed(self.random_state)

    def _init_params(self, X=None, diag_uniform_dist = (.95, .99), output_hmm_params=True):
        """
        Function to initialize HMM parameters. Can do so using kmeans++, randomly or deterministic.

        Parameters
        ----------
        diag_uniform_dist: 1D-array
            The lower and upper bounds of uniform distribution to sample init from.

        Attributes
        -------
        self.T: N X N matrix of transition probabilities
        self.delta: 1 X N vector of initial probabilities
        self.mu: 1 X N vector of state dependent means
        self.std: 1 X N vector of state dependent STDs

        """
        if self.init == 'random':
            # Transition probabilities
            trans_prob = np.diag(np.random.uniform(low=diag_uniform_dist[0], high=diag_uniform_dist[1],
                                                   size=self.n_states))  # Sample diag as uniform dist
            remaining_row_nums = (1 - np.diag(trans_prob)) / (
                        self.n_states - 1)  # Spread the remaining mass evenly onto remaining row values
            trans_prob += remaining_row_nums.reshape(-1, 1)  # Add this to the uniform diagonal matrix
            np.fill_diagonal(trans_prob, trans_prob.diagonal() - remaining_row_nums)  # And subtract these numbers from the diagonal so it remains uniform

            # initial distribution
            init_dist = np.ones(self.n_states) / np.sum(self.n_states)  # initial distribution 1 X N vector

            # State dependent distributions
            mu = np.random.uniform(low=-0.05, high=0.15, size=self.n_states)  # np.random.rand(self.n_states)
            std = np.random.uniform(low=0.01, high=0.3, size=self.n_states)  # np.random.rand(self.n_states) #

            self.tpm = trans_prob
            self.start_proba = init_dist
            self.mu = mu
            self.std = std

        elif self.init == "deterministic":  # TODO deprecate
            # Init theta as zeros and sample state seq from uniform dist
            self.theta = np.zeros(shape=(self.n_features, self.n_states))  # Init as empty matrix
            state_index = np.arange(start=0, stop=self.n_states, step=1, dtype=int)  # Array of possible states
            state_seq = np.random.choice(a=state_index, size=len(X))  # Sample sequence uniformly
            self.state_seq = state_seq

            trans_prob = np.zeros((2, 2))
            trans_prob[0, 0] = 0.7
            trans_prob[0, 1] = 0.3
            trans_prob[1, 0] = 0.2
            trans_prob[1, 1] = 0.8
            init_dist = np.array([0.2, 0.8])  # initial distribution 1 X N vector
            mu = [-0.05, 0.1]
            std = [0.2, 0.1]

            self.tpm = trans_prob
            self.start_proba = init_dist
            self.mu = mu
            self.std = std

    def emission_probs(self, X):
        """
        Computes the probability distribution p(x) given an observation sequence X
        The calculation will return a T X N matrix

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Time series of data

        Returns
        ----------
        probs : ndarray of shape (n_samples, n_states)
            Probability for sampling from a particular state distribution
        log_probs : ndarray of shape (n_samples, n_states)
            Log probability for sampling from a particular state distribution
        """

        T = len(X)
        log_probs = np.zeros((T, self.n_states))  # Init T X N matrix
        probs = np.zeros((T, self.n_states))

        # For all states evaluate the density function
        for j in range(self.n_states):
            if self.std[j] == np.nan or self.std[j] == 0. or self.std[j] < 0.:
                continue  # If std has non-standard value keep the probs at zero and go to next loop
            log_probs[:, j] = stats.norm.logpdf(X, loc=self.mu[j], scale=self.std[j])

        probs = np.exp(log_probs)

        self.emission_probs_ = probs
        self.log_emission_probs_ = log_probs

        return probs, log_probs

    def fit_predict(self, X, n_preds=15, verbose=False):
        """
        Fit model, then decode states and make n predictions.
        Wraps .fit(), .decode() and .predict_proba() into one method.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Data to be fitted.
        n_preds : int, default=15
            Number of time steps to look forward from current time

        Returns
        -------

        """
        self.fit(X, sort_state_seq=True, verbose=False,)
        if self.is_fitted == False:  # Check if model is fitted
            max_iter = self.max_iter
            self.max_iter = max_iter * 2  # Double amount of iterations
            self.fit(X)  # Try fitting again
            self.max_iter = max_iter  # Reset max_iter back to user-input
            if self.is_fitted == False and verbose == True:
                print(f'NOT FITTED -- mu {self.mu} -- std {self.std} -- tpm {np.diag(self.tpm)}')

        if self.type == 'mle':
            state_sequence = self.decode(X)  # 1D-array with most likely state sequence
        elif self.type == 'jump':
            state_sequence = self.state_seq

        # Posterior probability of being in state j at time t
        self.emission_probs_, self.log_emission_probs_ = self.emission_probs(X)
        posteriors = self.predict_proba(n_preds)  # 2-D array of shape (n_preds, n_states)

        return state_sequence, posteriors

    def sample(self, n_samples, n_sequences=1, hmm_params=None):
        '''
        Sample states from a fitted Hidden Markov Model.

        Parameters
        ----------
        n_samples : int
            Amount of samples to generate
        n_sequences : int, default=1
            Number of independent sequences to sample from, e.g. if n_samples=100 and n_sequences=3
            then 3 different sequences of length 100 are sampled
        hmm_params : dict, default=None
            hmm model parameters to sample from. If None and model is fitted it will use fitted parameters.
            To manually set params, create a dict with 'mu', 'std', 'tpm' and 'stationary distribution' as kwds
            and values ndarrays.

        Returns
        -------
        samples : ndarray of shape (n_samples,)
            Outputs the generated samples of size n_samples
        sample_states : ndarray of shape (n_samples, n_sequences)
            Outputs sampled states
        '''
        if hmm_params == None:
            mu = self.mu
            std = self.std
            tpm = self.tpm
            stationary_dist = self.stationary_dist
        else:
            mu = hmm_params['mu']
            std = hmm_params['std']
            tpm = hmm_params['tpm']
            stationary_dist = hmm_params['stationary_dist']


        state_index = np.arange(start=0, stop=self.n_states, step=1, dtype=np.int32)  # Array of possible states
        sample_states = np.zeros(shape=(n_samples, n_sequences), dtype=np.int32) # Init sample vector
        samples = np.zeros(shape=(n_samples, n_sequences))  # Init sample vector

        for seq in range(n_sequences):
            sample_states[0, seq] = np.random.choice(a=state_index, size=1, p=stationary_dist)

            for t in range(1, n_samples):
                # Each new state is chosen using the transition probs corresponding to the previous state sojourn.
                sample_states[t, seq] = np.random.choice(a=state_index, size=1, p=tpm[sample_states[t - 1, seq], :])

            try: # Prevents program from crashing due to "Domain errors"
                samples[:, seq] = stats.norm.rvs(loc=mu[sample_states[:, seq]], scale=std[sample_states[:, seq]], size=n_samples)
            except Exception:
                print(sys.exc_info()[0])
                print(f'Error occurrred with params: MU {self.mu} -- STD {self.std}, -- TPM {self.tpm.ravel()}')

        if n_sequences == 1:
            sample_states = sample_states[:, 0]
            samples = samples[:, 0]

        return samples, sample_states

    def decode(self, X):
        """
        Function to output the most likely sequence of states given an observation sequence.

        Parameters
        ----------
        X : ndarray of shape (n_samples,)
            Times series of data

        Returns
        ----------
        state_preds : ndarray of shape (n_samples,)
            Predicted sequence of states with length of the inputted time series.
        """
        state_preds = self._viterbi(X)
        return state_preds

    def predict_proba(self, n_preds=15):
        """
        Compute the probability P(St+h = i | X^T = x^T).
        Calculates the probability of being in state i at future time step h given a specific observation sequence up untill time T.

        Parameters
        ----------
        n_preds : int, default=15
            Number of time steps to look forward from current time

        Returns
        ----------
        state_preds : ndarray of shape (n_states, n_preds)
            Output the probability of being in state i at time t+h
        """

        llk, log_alphas = self._log_forward_proba() # Compute scaled log-likelihood

        # Same as posterior at terminal time. Subtracting llk to get unscaled probability.
        # Since backward_proba = 1 at last time step, this is omitted
        state_pred_t = np.exp(log_alphas[-1, :] - llk)

        state_preds = np.zeros(shape=(n_preds, self.n_states))  # Init matrix of predictions
        for t in range(n_preds):
            state_preds[t] = state_pred_t = state_pred_t @ self.tpm

        return state_preds

    def get_stationary_dist(self, tpm):
        """
        Outputs the stationary distribution of the fitted model.

        The stationary distributions corresponds to the largest eigenvector of the
        transition probability matrix. Since all values in the TPM are bounded between 0 and 1,
        we know that the largest eigenvalue is 1, and that the eigenvectors will all be defined by real numbers.

        Computed by taking the eigenvector corresponding to the largest eigenvalue and scaling it to sum to 1.

        Returns
        -------
        stationary_dist: ndarray of shape (n_states,)
        """
        if tpm.ndim < 2:  # In cases where only 1 state is detected
            return np.array([1., 0.])

        # Computes right eigenvectors, thus transposing the TPM is necessary.
        eigvals, eigvecs = np.linalg.eig(tpm.T)
        eigvec = eigvecs[:, np.argmax(eigvals)]  # Get the eigenvector corresponding to largest eigenvalue

        stationary_dist = eigvec / eigvec.sum()
        return stationary_dist

    def _log_forward_proba(self):
        r"""
        Compute log forward probabilities in scaled form.

        Forward probability is essentially the joint probability of observing
        a state = i and observation sequences x^t=x_1...x_t, i.e. $P(S_t=i , X^t=x^t)$.
        Follows the method by Zucchini A.1.8 p 334.

        Returns
        -------
        log-likelihood : float
            log-likehood of given HMM parameters
        log of forward probabilities : ndarray of shape (n_samples, n_states)
            Array of the scaled log of forward probabilities at each time step.
        """
        n_obs, n_states = self.log_emission_probs_.shape
        log_alphas = np.zeros((n_obs, n_states))

        self.check_params_not_zero()

        # Do the pass in cython
        with np.errstate(divide='ignore'):
            hmm_cython.forward_proba(n_obs, n_states,
                                     np.log(self.start_proba),
                                     np.log(self.tpm),
                                     self.log_emission_probs_, log_alphas)
            return logsumexp(log_alphas[-1]), log_alphas  # log-likelihood and forward probabilities

    def _log_backward_proba(self):
        """
        Compute the log of backward probabilities in scaled form.
        Backward probabilities are the conditional probability of
        some observation at t+1 given the current state = i. Equivalent to P(X_t+1 = x_t+1 | S_t = i)

        Returns
        -------
        log of backward probabilities: ndarray of shape (n_samples, n_states)
        """
        n_obs, n_states = self.log_emission_probs_.shape
        log_betas = np.zeros((n_obs, n_states))

        # Do the pass in cython
        with np.errstate(divide='ignore'):
            hmm_cython.backward_proba(n_obs, n_states,
                                      np.log(self.start_proba),
                                      np.log(self.tpm),
                                      self.log_emission_probs_, log_betas)
            return log_betas

    def _viterbi(self, X):
        self.emission_probs_, self.log_emission_probs_ = self.emission_probs(X)
        n_obs, n_states = self.log_emission_probs_.shape
        state_sequence = hmm_cython.viterbi(n_obs, n_states,
                                            np.log(self.start_proba),
                                            np.log(self.tpm),
                                            self.log_emission_probs_)
        return state_sequence.astype(np.int32)

    def check_params_not_zero(self):
        if self.start_proba[0] == 0:
            self.start_proba[0] = 0.0000001
            self.start_proba[1] = 0.9999999
        elif self.start_proba[1] == 0:
            self.start_proba[1] = 0.0000001
            self.start_proba[0] = 0.9999999

        if self.tpm[0, 0] == 1.:
            self.tpm[0, 0] = 0.999999
            self.tpm[0, 1] = 0.000001
        elif self.tpm[1, 1] == 1.:
            self.tpm[1, 1] = 0.999999
            self.tpm[1, 0] = 0.000001

    def fit(self, X, get_hmm_params=True, sort_state_seq=True, verbose=False, feature_set='feature_set_2'):
        """
        fit model to data. Defined in respective child classes
        """
        pass

    def rolling_posteriors(self, X):
        """ Compute the posterior probability of being in state i at time T.

        Function should be used as part of rolling estimation when one is
        interested only in the final smoothing probability of the current window sample.
        """
        if not self.is_fitted is True:
            print('Warning. Trying to predict using an unfitted model.')

        # Use emission probs to do a forward pass
        # Note only forward probs are needed as backward probs at time T is 1 by definition.
        self.emission_probs(X)
        llk, log_alphas = self._log_forward_proba()

        # Probability of being in state i in time T
        posterior = np.exp(log_alphas[-1] - llk)
        return posterior

    def bac_score(self, X, y_true, verbose=False):
        """ Computes balanced accuracy score when true state sequence is known """
        if not self.is_fitted is True:
            bac = 0.
            return bac
        if self.type == 'mle' or self.type == 'sampler':
            y_pred = self.decode(X)
        elif self.type == 'jump':
            y_pred = self.state_seq
            y_true = y_true[(self.window_len[-1] - 1):]  # slice y_true to have same dim as y_pred

        conf_matrix = confusion_matrix(y_true, y_pred)
        keep_idx = conf_matrix.sum(axis=1) != 0  # Only keep non-zero rows
        conf_matrix = conf_matrix[keep_idx]

        tp = np.diag(conf_matrix)
        fn = conf_matrix.sum(axis=1) - tp
        tpr = tp / (tp + fn)
        bac = np.mean(tpr)

        if verbose is True:
            logical_1 = bac < 0.5
            logical_2 = conf_matrix.ndim > 1 and \
                        np.any(conf_matrix.sum(axis=1)==0)

            if logical_1 or logical_2:
                print(f'bac {bac} -- tpr {tpr}')
                print(conf_matrix)

        return bac


import numpy as np
from scipy import stats
from scipy.special import logsumexp
import matplotlib.pyplot as plt

from hmmpy.sampler import SampleHMM
from hmmpy.base import BaseHiddenMarkov
from hmmpy.mle import MLEHMM


class OnlineHMM(MLEHMM, BaseHiddenMarkov):

    def __init__(self, n_states: int = 2, init: str = 'random', max_iter: int = 100, tol: float = 1e-6,
                 epochs: int = 10, random_state: int = 42):
        super().__init__(n_states, init, max_iter, tol, epochs, random_state)

    def _init_posteriors(self, X, forget_fac):
        self.emission_probs(X)
        llk, log_alphas = self._log_forward_proba()
        log_betas = self._log_backward_proba()
        log_posteriors = self.compute_log_posteriors(log_alphas, log_betas)  # 2-D array (n_samples, n_states)

        forget_factors = forget_fac**(np.arange(1, len(X)+1, 1)[::-1])
        posteriors = np.exp(log_posteriors) * forget_factors[:, np.newaxis]

        return posteriors

    def train_no_expsmoothing(self, X, n_init_obs=250, forget_fac=0.9925):
        self._init_params()

        #posteriors_exp = self._init_posteriors(X[:250], forget_fac=forget_fac)
        #X = X[250:]

        self.log_forward_proba = np.zeros(shape=(len(X), self.n_states))
        self.posteriors = np.zeros(shape=(len(X), self.n_states))
        self.rec = np.zeros(shape=(len(X), self.n_states))

        self.log_forward_proba[0] = np.log(self.start_proba) + stats.norm.logpdf(X[0], loc=self.mu, scale=self.std)
        self.posteriors[0] = np.exp(self.log_forward_proba[0] - logsumexp(self.log_forward_proba[0]))
        #self.rec[0] = forget_fac * posteriors_exp[-1] + (1 - forget_fac) * self.posteriors[0]

        for t in range(1, len(X)):

            log_tpm = np.log(self.tpm)
            log_proba = stats.norm.logpdf(X[t], loc=self.mu, scale=self.std)
            self.log_forward_proba[t] = logsumexp(log_tpm + self.log_forward_proba[t - 1], axis=0) + log_proba
            llk = logsumexp(self.log_forward_proba[t])

            self.posteriors[t] = np.exp(self.log_forward_proba[t] - llk)
            #self.rec[t] = forget_fac * self.rec[t - 1] + (1 - forget_fac) * self.posteriors[t]

            log_xi = np.zeros((2, 2))  # TODO move outside loop?
            for i in range(self.n_states):
                for j in range(self.n_states):
                    log_xi[i, j] = self.log_forward_proba[t - 1, i] + log_tpm[i, j] \
                                        + log_proba[j]

            normalizer = logsumexp(log_xi)
            log_xi = log_xi - normalizer
            xi = np.exp(log_xi)

            # TODO inefficient to sum everything in each loop
            sum1 = np.sum(self.posteriors[1:t], axis=0)
            sum2 = np.sum(self.posteriors[1:t+1], axis=0)

            if t > 2:
                self.tpm = (sum1 / sum2).reshape(-1, 1) * self.tpm \
                      + xi / sum2.reshape(-1, 1)

            self.tpm = self.tpm / self.tpm.sum(axis=1).reshape(-1, 1)

            print(self.tpm)
            print('-'*40)

            #self.tpm = (self.rec[t-1] / self.rec[t] * self.tpm.T).T + (self.trans_proba.T / self.rec[t]).T
            #self.tpm = self.tpm / np.sum(self.tpm, axis=1).reshape((-1, 1))

    def train(self, X, n_init_obs=500, forget_fac=0.9925):
        # Initialize mu, std and tpm
        self._init_params()

        # Set mu and std to their real values to test functionality
        self.mu = np.array([100, 200])
        self.std = np.array([10, 10])

        #prev_posteriors = self._init_posteriors(X[:n_init_obs], forget_fac=forget_fac)
        #X = X[n_init_obs:]

        # Init empty arrays for storing values
        self.log_forward_proba = np.zeros(shape=(len(X), self.n_states))
        self.posteriors = np.zeros(shape=(len(X), self.n_states))
        self.rec = np.zeros(shape=(len(X), self.n_states))  # Exponential forgetting factor

        # Compute initial values before entering loop
        self.log_forward_proba[0] = np.log(self.start_proba) + stats.norm.logpdf(X[0], loc=self.mu, scale=self.std)
        self.posteriors[0] = np.exp(self.log_forward_proba[0] - logsumexp(self.log_forward_proba[0]))
        self.rec[0] =  (1 - forget_fac) * self.posteriors[0]

        # Arrays for storing values used for plotting later
        tpm = np.zeros(shape=(len(X), self.n_states))
        mu = np.zeros(shape=(len(X), self.n_states))
        std = np.zeros(shape=(len(X), self.n_states))
        tpm[0] = np.diag(self.tpm)
        mu[0] = self.mu
        std[0] = self.std

        for t in range(1, len(X)):
            # Compute log forward probabilities and log-likelihood
            log_tpm = np.log(self.tpm)
            log_proba = stats.norm.logpdf(X[t], loc=self.mu, scale=self.std)
            self.log_forward_proba[t] = logsumexp(log_tpm + self.log_forward_proba[t - 1], axis=0) + log_proba
            llk = logsumexp(self.log_forward_proba[t])

            # Exponential transformation to get posterior and its current sum
            self.posteriors[t] = np.exp(self.log_forward_proba[t] - llk)
            self.rec[t] = forget_fac * self.rec[t - 1] + (1 - forget_fac) * self.posteriors[t]

            # Compute log probability of state shift at time t
            log_xi = np.zeros((2, 2))  # TODO move outside loop?
            for i in range(self.n_states):
                for j in range(self.n_states):
                    log_xi[i, j] = self.log_forward_proba[t - 1, i] + log_tpm[i, j] \
                                   + log_proba[j] - llk
            xi = np.exp(log_xi)

            # Compute tpm, mu and std based on Stenger (2001) formulas
            forget_frac = self.rec[t-1] / self.rec[t]
            self.tpm = forget_frac.reshape(-1, 1) * self.tpm \
                       + xi / self.rec[t].reshape(-1, 1)
            self.tpm = self.tpm / self.tpm.sum(axis=1).reshape(-1, 1)  # Make rows sum to 1
            self.mu = forget_frac * self.mu + self.posteriors[t] * X[t] / self.rec[t]
            var = forget_frac * np.square(self.std) + self.posteriors[t] * (X[t]-self.mu)**2 / self.rec[t]
            self.std = np.sqrt(var)

            tpm[t] = np.diag(self.tpm)
            mu[t] = self.mu
            std[t] = self.std
        return tpm, mu, std

if __name__ == '__main__':
    # Set hmm params equal to those of Stenger (2001)
    hmm_params = {'mu': np.array([100, 200]) ,
                  'std': np.sqrt(np.array([100, 100])),
                  'tpm': np.array([[0.9, 0.1],  # TODO figure out powers of vectors in python
                                   [0.2, 0.8]])
                  }
    # Sample observations from given hmm
    sampler = SampleHMM(n_states=2, random_state=42, hmm_params=hmm_params)
    X, viterbi_states, true_states = sampler.sample_with_viterbi(1000, 1)

    # Train online model
    model = OnlineHMM(n_states=2, init='random', random_state=42)
    tpm, mu, std = model.train(X, forget_fac=0.99)

    print(
        f'Means: {model.mu} - STD {model.std} - TPM {np.diag(model.tpm)} ')

    fig, ax = plt.subplots(nrows=4, ncols=2, figsize=(12,7), sharex=True)
    ax[0, 0].plot(tpm[:, 0], label='P(1)', )
    ax[0, 0].plot([sampler.tpm[0,0]]*len(tpm), label='True P1', ls='dashed')
    ax[0, 1].plot(tpm[:, 1], label='P(2)', )
    ax[0, 1].plot([sampler.tpm[1,1]]*len(tpm), label='True P2', ls='dashed')
    #ax[0].legend()

    ax[1, 0].plot(mu[:, 0])
    ax[1, 0].plot([sampler.mu[0]]*len(mu), ls='dashed')
    ax[1, 1].plot(mu[:, 1])
    ax[1, 1].plot([sampler.mu[1]] * len(mu), ls='dashed')

    ax[2, 0].plot(std[:, 0])  # label='True states', ls='dashed')
    ax[2, 0].plot([sampler.std[0]] * len(std), ls='dashed')
    ax[2, 1].plot(std[:, 1])  # label='True states', ls='dashed')
    ax[2, 1].plot([sampler.std[1]] * len(std), ls='dashed')

    ax[3, 0].plot(model.rec[:, 0])
    ax[3, 1].plot(model.rec[:, 1])

    plt.show()

    from hmmpy.utils.plotting.plot_hmm import plot_samples_states
    plot_samples_states(X, true_states)